#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath

try:
    import tomllib
except ModuleNotFoundError:
    raise ModuleNotFoundError('无法导入 "tomllib" 模块, 该模块仅 Python 3.11+ 版本支持.')

from .path_util import PathUtil


class TOMLUtil(object):
    """TOML 工具类, 仅提供 TOML 解析功能."""

    @staticmethod
    def loads(text: str):
        return tomllib.loads(text)

    @staticmethod
    def load(fp: 'FileDescriptorOrPath', encoding: str = 'utf-8'):
        with PathUtil.open(fp, 'r', encoding=encoding) as f:
            return tomllib.load(f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from importlib import metadata


class PackageUtil:
    """包工具类"""

    @staticmethod
    def get_version(package_name: str) -> str:
        """获取包的版本字符串

        Args:
            package_name (str): 包名

        Returns:
            _type_: 版本字符串
        """
        return metadata.version(package_name)


if __name__ == '__main__':
    print(PackageUtil.get_version('easy_pyoc'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from ..utils.object_util import ObjectUtil


class Magic(object):
    """魔法类，简单实现了 __str__、__repr__、__call__ 方法"""
    _str_exclude = {}
    _repr_exclude = {}
    _call_exclude = {}
    _str_include = {}
    _repr_include = {}
    _call_include = {}

    def __str__(self) -> str:
        return ObjectUtil.repr(self, self._str_exclude, self._str_include)

    def __repr__(self) -> str:
        return ObjectUtil.repr(self, self._repr_exclude, self._repr_include)

    def __call__(self) -> dict:
        return ObjectUtil.vars(self, self._call_exclude, self._call_include)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, Callable, Any
import json

if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath

from .path_util import PathUtil


class JSONUtil(object):

    @staticmethod
    def loads(
        text: str,
        *,
        cls: type[json.JSONDecoder] | None = None,
        object_hook: Callable[[dict[Any, Any]], Any] | None = None,
        parse_float: Callable[[str], Any] | None = None,
        parse_int: Callable[[str], Any] | None = None,
        parse_constant: Callable[[str], Any] | None = None,
        object_pairs_hook: Callable[[list[tuple[Any, Any]]], Any] | None = None,
        **kw: Any,
    ):
        """反序列化 JSON 字符串 *text* (*str*, *bytes*, 或 *bytearray*) 为 Python 对象.

        To use a custom `JSONDecoder` subclass, specify it with the `cls`
        kwarg; otherwise `JSONDecoder` is used.

        Args:
            text (str): JSON 字符串
            cls (type[json.JSONDecoder] | None, optional):
                `JSONDecoder` 子类. 默认为 None.
            object_hook (Callable[[dict[Any, Any]], Any] | None, optional):
                `object_hook` is an optional function that will be called with
                the result of any object literal decode (a `dict`). The return
                value of `object_hook` will be used instead of the `dict`.
                This feature can be used to implement custom decoders
                (e.g. JSON-RPC class hinting). 默认为 None.
            parse_float (Callable[[str], Any] | None, optional):
                `parse_float`, if specified, will be called with the string of
                every JSON float to be decoded. By default this is equivalent to
                float(num_str). This can be used to use another datatype or parser
                for JSON floats (e.g. decimal.Decimal). 默认为 None.
            parse_int (Callable[[str], Any] | None, optional):
                `parse_int`, if specified, will be called with the string of every
                JSON int to be decoded. By default this is equivalent to int(num_str).
                This can be used to use another datatype or parser for JSON integers
                (e.g. float). 默认为 None.
            parse_constant (Callable[[str], Any] | None, optional):
                `parse_constant`, if specified, will be called with one of the
                following strings: -Infinity, Infinity, NaN.
                This can be used to raise an exception if invalid JSON numbers
                are encountered. 默认为 None.
            object_pairs_hook (Callable[[list[tuple[Any, Any]]], Any] | None, optional):
                `object_pairs_hook` is an optional function that will be called with
                the result of any object literal decoded with an ordered list of pairs.
                The return value of `object_pairs_hook` will be used instead of the
                `dict`. This feature can be used to implement custom decoders.  If
                `object_hook` is also defined, the `object_pairs_hook` takes
                priority. 默认为 None.

        Returns:
            Any: Python 对象.
        """
        return json.loads(
            text,
            cls=cls,
            object_hook=object_hook,
            parse_float=parse_float,
            parse_int=parse_int,
            parse_constant=parse_constant,
            object_pairs_hook=object_pairs_hook,
            **kw,
        )

    @staticmethod
    def dumps(
        obj: Any,
        *,
        indent: int = 4,
        ensure_ascii: bool = False,
        default: Callable[[Any], Any] | None = None,
    ):
        return json.dumps(obj, indent=indent, ensure_ascii=ensure_ascii, default=default)

    @staticmethod
    def load(fp: 'FileDescriptorOrPath', *, encoding: str = 'utf-8'):
        with PathUtil.open(fp, 'r', encoding=encoding) as f:
            return json.load(f)

    @staticmethod
    def dump(
        obj: Any,
        fp: 'FileDescriptorOrPath',
        *,
        encoding: str = 'utf-8',
        indent: int = 4,
        ensure_ascii: bool = False,
        default: Callable[[Any], Any] | None = None,
    ):
        with PathUtil.open(fp, 'w', encoding=encoding) as f:
            json.dump(obj, f, indent=indent, ensure_ascii=ensure_ascii, default=default)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath

try:
    import yaml
except ModuleNotFoundError:
    raise ModuleNotFoundError('使用该模块前，请先安装 PyYAML 库：`pip install PyYAML`')

from .path_util import PathUtil


class YAMLUtil:
    """YAML 工具类"""

    @staticmethod
    def loads(text: str):
        """将 yaml 字符串转换为 dict"""
        return yaml.safe_load(text)

    @staticmethod
    def dumps(data: dict):
        """将 dict 转换为 yaml 字符串"""
        return yaml.safe_dump(data)

    @staticmethod
    def load(fp: 'FileDescriptorOrPath', encoding='utf-8'):
        """读取 yaml 文件"""
        with PathUtil.open(fp, 'r', encoding=encoding) as f:
            data = yaml.safe_load(f)
        return data

    @staticmethod
    def dump(fp: 'FileDescriptorOrPath', data: dict, encoding='utf-8'):
        """写入 yaml 文件"""
        with PathUtil.open(fp, 'w', encoding=encoding) as f:
            yaml.safe_dump(data, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Any, Iterable, Mapping, Callable, Optional
from threading import Thread, Event, current_thread, active_count


class ThreadUtil:

    tasks: dict[int, tuple[Thread, Event]] = {}
    """任务字典, key 为任务 id, value 为 (任务对象, 停止事件) """

    @staticmethod
    def get_current_name():
        return current_thread().name

    @staticmethod
    def get_current_id():
        return current_thread().ident

    @staticmethod
    def get_count():
        return active_count()

    @staticmethod
    def get_list():
        return enumerate()

    @staticmethod
    def execute_task(
        target: Callable[[Event], None],
        args: Iterable[Any] = (),
        kwargs: Optional[Mapping[str, Any]] = None,
        daemon: bool = True,
    ) -> Optional[int]:
        """执行一个任务, 该任务应该通过 stop_flag.is_set() 来判断是否需要停止.

        Args:
            target (Callable[[Event], None]): 运行的函数, 第一个位置参数固定为 stop_flag.
            args (Iterable[Any], optional): 位置参数. 默认为 ().
            kwargs (Optional[Mapping[str, Any]], optional): 关键字参数. 默认为 None.
            daemon (bool, optional): 是否是守护线程. 默认为 True.

        Returns:
            Optional[int]: 任务 id
        """
        # 清除未活动的任务
        need_clear = [task_id for task_id, (task, _) in ThreadUtil.tasks.items() if not task.is_alive()]
        for task_id in need_clear:
            ThreadUtil.tasks.pop(task_id, None)

        stop_flag = Event()
        stop_flag.set()

        task = Thread(target=target, args=(stop_flag, *args), kwargs=kwargs, daemon=daemon)
        task.start()

        task_id = task.ident
        ThreadUtil.tasks[task_id] = (task, stop_flag)

        return task_id

    @staticmethod
    def chancel_task(task_id: int):
        """取消一个任务, 并调用 stop_flag.clear()

        Args:
            task_id (int): 任务 id

        Raises:
            ValueError: 任务不存在或已结束
        """
        (task, stop_flag) = ThreadUtil.tasks.get(task_id, (None, None))

        if not task:
            raise ValueError(f'任务 "{task_id}" 不存在')

        stop_flag.clear()
        task.join()
        ThreadUtil.tasks.pop(task_id, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, Optional, Union
from xml.dom.minidom import Document, parseString
from xml.etree.ElementTree import (
    ParseError,
    XMLParser,
    Element,
    fromstring,
    tostring,
)

if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath

from .path_util import PathUtil


class XMLUtil(object):
    """XML 工具类"""

    attr_prefix = '@'
    """属性节点前缀"""
    content_key = '#text'
    """内容节点键"""

    @staticmethod
    def auto_type(text: Optional[str]) -> Optional[Union[str, int, float, bool]]:
        """自动类型转换, 用于将 xml 字符串转换为字典时, 自动将字符串转换为数字或布尔类型"""
        if text is None:
            return None
        elif text.isdigit():
            return int(text)
        elif text.replace('.', '', 1).isdigit():
            return float(text)
        elif text.lower() in ['true', 'false']:
            return text.lower() == 'true'
        else:
            return text

    @staticmethod
    def loads(xml: str, *, parser: Optional[XMLParser] = None) -> dict:
        """将 xml 字符串转换为字典

        Args:
            xml (str): xml 字符串
            parser (XMLParser, optional): xml 解析器, 当为 `None` 时, 使用标准解析器. 默认为 `None`.

        Returns:
            dict: 解析后的字典
        """
        return XMLUtil.xml_to_dict(xml, parser=parser)

    @staticmethod
    def dumps(data: dict, *, tag: Optional[str] = None, deep: int = -1, encoding: str = 'unicode', method: str = 'xml') -> str | bytes:
        """将字典转换为 xml 字符串

        如果 encoding 为 `unicode`, 则返回字符串, 否则返回字节串

        Args:
            data (dict): 字典数据
            tag (Optional[str], optional): 根节点标签. 默认为 `root`.
            deep (int, optional): 递归深度, 当 deep 为 `-1` 时, 则不限制递归深度. 默认为 `-1`.
            encoding (str, optional): 编码格式. 默认为 `unicode`.
            method (str, optional): 输出方法, 可选 `xml` | `html` | `text` | `c14n`. 默认为 `xml`.

        Returns:
            str | bytes: xml 字符（节）串
        """
        return XMLUtil.dict_to_xml(data, tag=tag, deep=deep, encoding=encoding, method=method)

    @staticmethod
    def load(fp: 'FileDescriptorOrPath', *, parser: Optional[XMLParser] = None, encoding: str = 'utf-8') -> dict:
        """读取文件 xml 文件并转换为字典

        Args:
            fp (FileDescriptorOrPath): 文件路径
            parser (XMLParser, optional): xml 解析器, 当为 `None` 时, 使用标准解析器. 默认为 `None`.
            encoding (str, optional): 文件编码格式. 默认为 `utf-8`.

        Returns:
            dict: 解析后的字典
        """
        with PathUtil.open(fp, 'r', encoding=encoding) as f:
            return XMLUtil.xml_to_dict(f.read(), parser=parser)

    @staticmethod
    def dump(fp: 'FileDescriptorOrPath', data: dict, *, tag: Optional[str] = None, xml_encoding: str = 'unicode', file_encoding: str = 'utf-8', deep: int = -1, method: str = 'xml') -> None:
        """将字典写入文件 xml 文件

        Args:
            fp (FileDescriptorOrPath): 文件路径
            data (dict): 字典数据
            xml_encoding (str, optional): xml 编码格式. 默认为 `unicode`.
            file_encoding (str, optional): 文件编码格式. 默认为 `utf-8`.
            tag (Optional[str], optional): 根节点标签. 默认为 `root`.
            deep (int, optional): 递归深度, 当 deep 为 `-1` 时, 则不限制递归深度. 默认为 `-1`.
            method (str, optional): 输出方法, 可选 `xml` | `html` | `text` | `c14n`. 默认为 `xml`.
        """
        with PathUtil.open(fp, 'w', encoding=file_encoding) as f:
            f.write(XMLUtil.dict_to_xml(data, tag=tag, deep=deep, encoding=xml_encoding, method=method))

    @staticmethod
    def toprettyxml(xml: str | bytes, *, indent: str = '  ', new_line: str = '\n', encoding: Optional[str] = None, standalone: Optional[bool] = None, parser: Optional[XMLParser] = None) -> str | bytes:
        """格式化 xml 字符（节）串

        Args:
            xml (str | bytes): xml 字符串或字节串.
            indent (str, optional): 缩进字符. 默认为 `  `.
            new_line (str, optional): 换行符. 默认为 `\n`.
            encoding (str, optional): 编码格式. 默认为 `None`.
            standalone (Optional[bool], optional):
            parser (XMLParser, optional): xml 解析器, 当为 `None` 时, 使用标准解析器. 默认为 `None`.

            str | bytes: 格式化后的 xml 字符（节）串
        """
        return parseString(xml, parser=parser).toprettyxml(indent, new_line, encoding, standalone)

    @staticmethod
    def element_to_dict(element: Element) -> dict:
        """将 xml.etree.ElementTree.Element 转换为字典

        Args:
            element (Element): xml.etree.ElementTree.Element 对象

        Returns:
            dict: 解析后的字典
        """
        result, text, tail = ({element.tag: {}}), element.text or '', ''

        # 设置属性
        for k, v in element.attrib.items():
            result[element.tag][f'{XMLUtil.attr_prefix}{k}'] = XMLUtil.auto_type(v)

        # 设置子节点
        for child in element:
            tail += child.tail or ''

            if child.tag in result[element.tag]:
                if isinstance(result[element.tag][child.tag], list):
                    result[element.tag][child.tag].append(XMLUtil.element_to_dict(child)[child.tag])
                else:
                    result[element.tag][child.tag] = [
                        result[element.tag][child.tag],
                        XMLUtil.element_to_dict(child)[child.tag],
                    ]
            else:
                result[element.tag].update(XMLUtil.element_to_dict(child))

        # 设置文本
        if not result[element.tag]:
            result[element.tag] = XMLUtil.auto_type((element.text or '').strip() or None)
        else:
            if text.lstrip() and tail.rstrip():
                result[element.tag][XMLUtil.content_key] = text.lstrip() + tail.rstrip()
            elif text.strip():
                result[element.tag][XMLUtil.content_key] = text.strip()
            elif tail.strip():
                result[element.tag][XMLUtil.content_key] = tail.strip()
        return result

    @staticmethod
    def xml_to_element(xml: str, *, parser: Optional[XMLParser] = None) -> Element:
        """将 xml 字符串转换为 xml.etree.ElementTree.Element 对象

        Args:
            xml (str): xml 字符串
            parser (XMLParser, optional): xml 解析器, 当为 `None` 时, 使用标准解析器. 默认为 `None`.

        Returns:
            Element: xml.etree.ElementTree.Element 对象
        """
        try:
            return fromstring(xml, parser)
        except ParseError:
            raise ValueError('XML 格式有误')

    @staticmethod
    def xml_to_dict(xml: str, *, parser: Optional[XMLParser] = None) -> dict:
        """将 xml 字符串转换为字典

        Args:
            xml (str): xml 字符串
            parser (XMLParser, optional): xml 解析器, 当为 `None` 时, 使用标准解析器. 默认为 `None`.

        Returns:
            dict: 解析后的字典
        """
        return XMLUtil.element_to_dict(
            XMLUtil.xml_to_element(
                xml,
                parser=parser,
            ),
        )

    @staticmethod
    def element_to_xml(element: Element, *, encoding: str = 'unicode', method: str = 'xml', xml_declaration: bool = False, short_empty_elements: bool = True, default_namespace: Optional[str] = None) -> str | bytes:
        """将 xml.etree.ElementTree.Element 对象转换为 xml 字符串

        如果 encoding 为 `unicode`, 则返回字符串, 否则返回字节串

        Args:
            element (Element): xml.etree.ElementTree.Element 对象
            encoding (str, optional): 编码格式. 默认为 `unicode`.
            method (str, optional): 输出方法, 可选 `xml` | `html` | `text` | `c14n`. 默认为 `xml`.
            xml_declaration (bool, optional): 是否输出 xml 声明. 默认为 `False`.
            short_empty_elements (bool, optional): 是否压缩空元素. 默认为 `True`.
            default_namespace (Optional[str], optional): 默认命名空间. 默认为 `None`.

        Returns:
            str | bytes: xml 字符（节）串
        """
        return tostring(element, encoding, method, xml_declaration=xml_declaration, default_namespace=default_namespace, short_empty_elements=short_empty_elements)

    @staticmethod
    def dict_to_element(data: dict, *, tag: Optional[str] = None, deep: int = -1) -> Element:
        """将字典转换为 xml.etree.ElementTree.Element 对象

        Args:
            data (dict): 字典数据
            tag (Optional[str], optional): 根节点标签. 默认为 `root`.
            deep (int, optional): 递归深度, 当 deep 为 `-1` 时, 则不限制递归深度. 默认为 `-1`.

        Returns:
            Element: xml.etree.ElementTree.Element 对象
        """
        element = Element(tag)

        if deep == 0:
            element.text = str(data)
            return element

        if isinstance(data, dict):
            for k, v in data.items():
                if k.startswith(XMLUtil.attr_prefix):
                    element.attrib[k[len(XMLUtil.attr_prefix):]] = str(v)
                elif k == XMLUtil.content_key:
                    element.text = str(v)
                elif isinstance(v, list):
                    for item in v:
                        child = XMLUtil.dict_to_element(item, tag=k, deep=deep - 1)
                        element.append(child)
                else:
                    child = XMLUtil.dict_to_element(v, tag=k, deep=deep - 1)
                    element.append(child)
        elif isinstance(data, list):
            for item in data:
                child = XMLUtil.dict_to_element(item, tag=tag, deep=deep - 1)
                element.append(child)
        elif data is not None:
            element.text = str(data)

        return element

    @staticmethod
    def dict_to_xml(data: dict, *, tag: Optional[str] = None, deep: int = -1, encoding: str = 'unicode', method: str = 'xml', xml_declaration: bool = False, short_empty_elements: bool = True, default_namespace: Optional[str] = None) -> str | bytes:
        """将字典转换为 xml 字符串

        如果 encoding 为 `unicode`, 则返回字符串, 否则返回字节串

        Args:
            data (dict): 字典数据
            tag (Optional[str], optional): 根节点标签. 默认为 `root`.
            deep (int, optional): 递归深度, 当 deep 为 `-1` 时, 则不限制递归深度. 默认为 `-1`.
            encoding (str, optional): 编码格式. 默认为 `unicode`.
            method (str, optional): 输出方法, 可选 `xml` | `html` | `text` | `c14n`. 默认为 `xml`.
            xml_declaration (bool, optional): 是否输出 xml 声明. 默认为 `False`.
            short_empty_elements (bool, optional): 是否压缩空元素. 默认为 `True`.
            default_namespace (Optional[str], optional): 默认命名空间. 默认为 `None`.

        Returns:
            str | bytes: xml 字符（节）串
        """
        return XMLUtil.element_to_xml(
            XMLUtil.dict_to_element(
                data,
                tag=tag,
                deep=deep,
            ),
            encoding=encoding,
            method=method,
            xml_declaration=xml_declaration,
            short_empty_elements=short_empty_elements,
            default_namespace=default_namespace,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .classes.logger import Logger as Logger
from .classes.magic import Magic as Magic
from .classes.config import Config as Config
from .classes.color import AnsiColor as AnsiColor

from .sock.server_socket import ServerSocket as ServerSocket
from .sock.client_socket import ClientSocket as ClientSocket

from .utils import not_this_module
from .utils.crc_util import CRCUtil as CRCUtil
from .utils.datetime_util import DateTimeUtil as DateTimeUtil
from .utils.json_util import JSONUtil as JSONUtil
from .utils.knx_util import KNXUtil as KNXUtil
from .utils.network_util import NetworkUtil as NetworkUtil
from .utils.object_util import ObjectUtil as ObjectUtil
from .utils.path_util import PathUtil as PathUtil
from .utils.string_util import StringUtil as StringUtil
from .utils.xml_util import XMLUtil as XMLUtil
from .utils.thread_util import ThreadUtil as ThreadUtil
from .utils.func_util import FuncUtil as FuncUtil
from .utils.package_util import PackageUtil as PackageUtil
from .utils.exception_util import ExceptionUtil as ExceptionUtil
from .utils.cr4_util import CR4Util as CR4Util

try:
    from .utils.flask_util import FlaskUtil as FlaskUtil
except ImportError as e:
    FlaskUtil = not_this_module('FlaskUtil', e.msg)

try:
    from .utils.toml_util import TOMLUtil as TOMLUtil
except ImportError as e:
    TOMLUtil = not_this_module('TOMLUtil', e.msg)

try:
    from .utils.yaml_util import YAMLUtil as YAMLUtil
except ImportError as e:
    YAMLUtil = not_this_module('YAMLUtil', e.msg)


__version__ = PackageUtil.get_version('easy_pyoc')
__author__ = 'one-ccs'
__email__ = 'one-ccs@foxmail.com'

__all__ = [
    'Logger',
    'Magic',
    'Config',
    'AnsiColor',

    'ServerSocket',
    'ClientSocket',

    'CRCUtil',
    'DateTimeUtil',
    'FlaskUtil',
    'JSONUtil',
    'KNXUtil',
    'NetworkUtil',
    'ObjectUtil',
    'PathUtil',
    'StringUtil',
    'TOMLUtil',
    'XMLUtil',
    'YAMLUtil',
    'ThreadUtil',
    'FuncUtil',
    'PackageUtil',
    'ExceptionUtil',
    'CR4Util',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Callable, Literal
from threading import Thread
from multiprocessing import Process
from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE
import socket

from ..classes.logger import Logger


class ServerSocket():
    def __init__(
            self,
            *,
            protocol: Literal['TCP', 'UDP', 'MULTICAST'],
            bind: tuple[str, int],
            group: str | None = None,
            on_recv: Callable[[bytes, tuple[str, int], Callable[[bytes], int]], None],
        ):
        """服务端套接字

        在新线程中创建 TCP/UDP/MULTICAST 协议的服务端套接字，接收客户端的
        连接请求或数据，并调用 on_recv 回调函数处理数据。

        TCP 断开连接的情况：

        - TCP 正常断开
            + 客户端主动断开连接
            + 通信期间正常交换数据 (若服务端返回了响应, 则客户端应该接收响应)
        - TCP 连接已重置
            + 客户端主动断开连接
            + 服务端返回了响应，但客户端未接收
        - TCP 连接已终止
            + 未通信完毕就已经断开了连接

        Args:
            protocol (str): 协议
            bind (tuple[str, int]): 绑定的地址, 端口为 `0` 时随机分配端口, 注: 当多网卡, 且 ip 为 "0.0.0.0" 时, 有可能接收不到数据.
            group (tuple[str, int] | None, optional): 组播地址, 仅在协议类型为 "MULTICAST" 时有效. 默认为 None.
            on_recv (Callable, optional): 接收到数据时的回调函数, 参数为 (data: bytes, client_name: str, send_back: Callable[[bytes], int]). 默认为 None.

        Raises:
            ValueError: 无效的协议类型, 应为 [TCP, UDP, MULTICAST]
            ValueError: 组播协议必须指定组播地址
            ValueError: 协议类型为非 "MULTICAST" 时请勿设置 group 参数
            ValueError: 绑定地址 (bind[0]) 不能为空
            ValueError: 无效的端口号, 应为 [1-65535]
        Examples:
        """
        self.__active = False
        self.__socked = False

        if protocol not in ['TCP', 'UDP', 'MULTICAST']:
            raise ValueError(f'ServerSocket 无效的协议类型 "{protocol}"')
        if protocol == 'MULTICAST' and not group:
            raise ValueError(f'ServerSocket 组播协议必须指定组播地址')
        if protocol != 'MULTICAST' and group:
            raise ValueError(f'ServerSocket 协议类型为 "{protocol}" 时请勿设置 group 参数')
        if not bind[0]:
            raise ValueError(f'ServerSocket 绑定地址不能为空')
        if bind[1] < 0 or bind[1] > 65535:
            raise ValueError(f'ServerSocket 无效的端口号 "{bind[1]}"')
        if not callable(on_recv):
            raise ValueError(f'ServerSocket on_recv 参数必须为可调用对象')

        self.logger = Logger()
        self.protocol = protocol
        self.bind = bind
        self.group = group
        self.on_recv = on_recv
        self.sock: socket.socket | None = None
        self.tcp_sub_socks: list[socket.socket] = []
        self.thread: Thread | Process | None = None

    def __str__(self) -> str:
        if self.protocol == 'MULTICAST':
            return f'ServerSocket({self.protocol}, bind {self.bind[0]}:{self.bind[1]}, group {self.group})'
        return f'ServerSocket({self.protocol}, bind {self.bind[0]}:{self.bind[1]})'

    def __del__(self) -> None:
        self.close()

    def __create_socket(self) -> None:
        if self.__socked:
            return True

        try:
            match self.protocol:
                case 'TCP':
                    self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    self.sock.bind(self.bind)
                    self.sock.listen(10)
                case 'UDP':
                    self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    self.sock.bind(self.bind)
                case 'MULTICAST':
                    self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    self.sock.bind(self.bind)
                    self.sock.setsockopt(
                        socket.IPPROTO_IP,
                        socket.IP_ADD_MEMBERSHIP,
                        socket.inet_aton(self.group) + socket.inet_aton(self.bind[0]),
                    )
            self.bind = self.sock.getsockname()
            self.__socked = True
        except Exception as e:
            self.logger.error(f'{self} 创建失败: \n{e}')

        return self.__socked

    def __send_back(self, client_addr: tuple[str, int], client_sock: socket.socket | None = None) -> Callable:
        def send_back(data: bytes):
            self.logger.debug(f'{self} 向 {client_addr} 返回数据: {data}')

            if self.protocol == 'TCP':
                return client_sock.sendto(data, client_addr)
            return self.sock.sendto(data, client_addr)

        return send_back

    def __tcp_sub_thread(self, client_sock: socket.socket, client_addr: tuple[str, int]) -> None:
        while self.is_active():
            try:
                if not (data := client_sock.recv(1024)):
                    self.logger.debug(f'{self} TCP 子线程 {client_addr} 正常断开')
                    break

                self.logger.debug(f'{self} TCP 子线程 {client_addr} 接收到数据: {data}')
                try:
                    self.on_recv(data, client_addr, self.__send_back(client_addr, client_sock))
                except Exception as e:
                    self.logger.error(f'{self} TCP 子线程 {client_addr} "on_recv" 回调函数发生异常: \n{e}')
            except ConnectionResetError:
                self.logger.debug(f'{self} TCP 子线程 {client_addr} 连接已重置')
                break
            except ConnectionAbortedError:
                self.logger.debug(f'{self} TCP 子线程 {client_addr} 连接已终止')
                break
            except Exception as e:
                if self.is_active():
                    self.logger.error(f'{self} TCP 子线程 {client_addr} 异常: \n{e}')
                break
        if self.is_active(): # 断开或异常断开
            self.tcp_sub_socks.remove(client_sock)
            client_sock.close()

    def __main_thread(self) -> None:
        self.__active = True

        while self.is_active():
            try:
                if self.protocol == 'TCP':
                    client_sock, client_addr = self.sock.accept()
                    self.logger.debug(f'{self} 与 {client_addr} 建立 TCP 连接')
                    self.tcp_sub_socks.append(client_sock)
                    Thread(target=self.__tcp_sub_thread, args=(client_sock, client_addr), daemon=True).start()
                else:
                    data, client_addr = self.sock.recvfrom(1024)

                    self.logger.debug(f'{self} 收到 {client_addr} 的数据: {data}')
                    try:
                        self.on_recv(data, client_addr, self.__send_back(client_addr))
                    except Exception as e:
                        self.logger.error(f'{self} 主线程 "on_recv" 回调函数发生异常: \n{e}')
            except Exception as e:
                if self.is_active():
                    self.logger.error(f'{self} 主线程异常 : \n{e}')
                    break

    def getpeername(self) -> tuple[str | None, int | None]:
        """返回套接字连接到的远程地址。"""
        if self.__create_socket():
            return self.sock.getpeername()
        return (None, None)

    def getsockname(self) -> tuple[str | None, int | None]:
        """返回套接字本身的地址。"""
        if self.__create_socket():
            return self.sock.getsockname()
        return (None, None)

    def send(self, data: bytes, client_addr: tuple[str, int]) -> int:
        """向指定客户端发送数据

        Args:
            data (bytes): 要发送的数据
            client_addr (tuple[str, int]): 客户端地址

        Returns:
            int: 实际发送的字节数, 为 -1 表示 socket 未建立或已关闭
        """
        if self.__create_socket():
            if self.protocol == 'TCP':
                for client_sock in self.tcp_sub_socks:
                    if client_sock.getpeername() == client_addr:
                        return client_sock.sendall(data)
                return 0
            return self.sock.sendto(data, client_addr)
        return -1

    def start(self, is_process: bool = False) -> bool:
        """启动服务端

        将在新线程中运行，直到调用 close() 关闭，TCP 协议下会创建子线程处理 TCP 连接

        Args:
            is_process (bool, optional): 是否以子进程运行. 默认为 False.

        Returns:
            bool: 是否启动成功
        """
        if not self.thread and self.__create_socket():
            if is_process:
                self.thread = Process(target=self.__main_thread, daemon=True)
            else:
                self.thread = Thread(target=self.__main_thread, daemon=True)
            self.thread.start()
            return True

        return self.is_active()

    def close(self) -> bool:
        """关闭服务端

        Returns:
            bool: 是否关闭成功
        """
        if self.__socked:
            try:
                self.__active = False
                if self.protocol == 'TCP':
                    for client_sock in self.tcp_sub_socks:
                        client_sock.shutdown(socket.SHUT_RDWR)
                        client_sock.close()
                    self.tcp_sub_socks.clear()
                try:
                    # windows 平台 TCP 无需 shutdown (此时 shutdown 会导致 10057 错误)
                    # linux 平台非 TCP shutdown 会报 107 错误
                    self.sock.shutdown(socket.SHUT_RDWR)
                except OSError as e:
                    if e.errno != 107 and e.errno != 10057:
                        raise e
                self.sock.close()
                if isinstance(self.thread, Process):
                    self.thread.terminate()
                else:
                    self.thread.join()

                self.__socked = False
                self.logger.debug(f'{self} 已关闭')
            except Exception as e:
                self.logger.error(f'{self} 关闭失败: \n{e}')

        return self.__socked

    def is_active(self) -> bool:
        """返回服务端是否处于活动状态

        Returns:
            bool: 是否处于活动状态
        """
        return self.__socked and self.__active



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from unittest.mock import MagicMock, call

from easy_pyoc import FuncUtil


def test_is_args():
    def test_func(a, b, c=10):
        pass

    assert FuncUtil.is_arg(test_func, 'a')
    assert FuncUtil.is_arg(test_func, 'b')
    assert not FuncUtil.is_arg(test_func, 'c')
    assert not FuncUtil.is_arg(test_func, 'd')

def test_is_kwarg():
    def test_func(a, b, c=10):
        pass

    assert not FuncUtil.is_kwarg(test_func, 'd')
    assert not FuncUtil.is_kwarg(test_func, 'a')
    assert not FuncUtil.is_kwarg(test_func, 'b')
    assert FuncUtil.is_kwarg(test_func, 'c')


def test_log_decorator():
    logger = MagicMock()

    @FuncUtil.log(logger=logger)
    def test_func(a, b):
        return a + b

    result = test_func(1, 2)
    assert result == 3
    logger.assert_called_once_with('调用 "test_func" 函数, 位置参数: (1, 2)')


def test_catch_decorator_no_exception():
    logger = MagicMock()

    @FuncUtil.catch(logger=logger)
    def test_func(a, b):
        return a + b

    result = test_func(1, 2)
    assert result == 3
    logger.assert_not_called()


def test_catch_decorator_with_exception():
    logger = MagicMock()

    @FuncUtil.catch(logger=logger, is_raise=True)
    def test_func(a, b):
        raise ValueError('test error')

    with pytest.raises(ValueError, match='test error'):
        test_func(1, 2)
    logger.assert_called_once()


def test_catch_decorator_with_exception_callback():
    logger = MagicMock()
    on_except = MagicMock()

    e = ValueError('test error')

    @FuncUtil.catch(logger=logger, is_raise=True, on_except=on_except)
    def test_func(a, b):
        raise e

    with pytest.raises(ValueError):
        test_func(1, 2)
    logger.assert_called_once()
    on_except.assert_called_with(e)


def test_catch_decorator_with_exception_no_raise():
    logger = MagicMock()

    @FuncUtil.catch(logger=logger, is_raise=False)
    def test_func(a, b):
        raise ValueError('test error')

    result = test_func(1, 2)
    assert result is None
    logger.assert_called_once()


def test_hook_decorator_before():
    hook = MagicMock(return_value=...)

    @FuncUtil.hook(hook=hook)
    def test_func(a, b):
        return a + b

    result = test_func(1, 2)
    assert result is None
    hook.assert_called_with('before', None, 1, 2)


def test_hook_decorator_after():
    def custom_side_effect(mode, result, a, b):
        if mode == 'after':
            return 10

    hook = MagicMock(side_effect=custom_side_effect)

    @FuncUtil.hook(hook=hook)
    def test_func(a, b):
        return a + b

    result = test_func(1, 2)
    assert result == 10
    hook.assert_has_calls([call('before', None, 1, 2), call('after', 3, 1, 2)])


def test_hook_decorator_after_no_return():
    hook = MagicMock(return_value=None)

    @FuncUtil.hook(hook=hook)
    def test_func(a, b):
        return a + b

    result = test_func(1, 2)
    assert result == 3
    hook.assert_has_calls([call('before', None, 1, 2), call('after', 3, 1, 2)])


def test_call_ignore_extra_params():
    def test_func(a, b, *, c=10):
        return a + b + c

    result = FuncUtil.call(test_func, 1, 2, 3, 4, c=5)
    assert result == 8
    result = FuncUtil.call(test_func, 1, 2, 3, 4)
    assert result == 13


def test_call_ignore_extra_params_keyword_only():
    def test_func(*, a, b, c=10):
        return a + b + c

    result = FuncUtil.call(test_func, 1, 2, a=3, b=4, d=5)
    assert result == 17



#!/usr/bin/env python
# -*- coding: utf-8 -*-

class CRCUtil(object):
    """CRC 校验工具类"""

    def crc16_xmodem(data: bytes, polynomial: int = 0x1021) -> int:
        """计算给定数据的 CRC16-XMODEM 校验值。

        Args:
            data (bytes): 需要计算 CRC 的字节数据。
            polynomial (hexadecimal, optional): CRC16-XMODEM 多项式。默认为 `0x1021`。

        Returns:
            int: 计算得到的 CRC16-XMODEM 校验值。
        """
        crc = 0  # 初始值为0

        for byte in data:
            crc ^= byte << 8  # 将字节移到高位，并与当前CRC值异或

            for _ in range(8):  # 对每个字节的每一位进行处理
                if crc & 0x8000:  # 如果最高位为1
                    crc = (crc << 1) ^ polynomial  # 左移一位并与多项式异或
                else:
                    crc <<= 1  # 否则左移一位
                crc &= 0xFFFF  # 保持CRC为16位

        return crc

    def crc16_modbus(data: bytes, polynomial: int = 0xA001) -> int:
        """计算指定数据的 CRC16-MODBUS 校验值。

        Args:
            data (bytes): 需要计算 CRC 的字节数据。
            polynomial (_type_, optional): CRC16-MODBUS 多项式。默认为 `0xA001`。

        Returns:
            int: 计算得到的 CRC16-MODBUS 校验值。
        """
        crc = 0xFFFF

        for byte in data:
            crc ^= byte  # 异或字节转换为 CRC 的最低有效字节

            for _ in range(8):  # 对每个字节的每一位进行处理
                if (crc & 0x0001):  # 如果设置了 LSB
                    crc >>= 1  # 右移一位
                    crc ^= polynomial  # 与多项式异或
                else:
                    crc >>= 1  # 仅右移一位
                crc &= 0xFFFF  # 保持 CRC 为16位

        return crc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from re import compile
from string import ascii_letters, digits, hexdigits


class StringUtil(object):

    base_table = digits + ascii_letters
    """基础字符表"""

    @staticmethod
    def camel_to_snake(camel_str: str) -> str:
        """将驼峰形式命名的字符串转换为下划线形式"""
        pattern = compile(r'(?<!^)(?=[A-Z])')
        return pattern.sub('_', camel_str).lower()

    @staticmethod
    def snake_to_camel(snake_str: str) -> str:
        """将下划线形式命名的字符串转换为驼峰形式"""
        if snake_str[0] == '_':
            components = snake_str.split('_')
            return '_' + ''.join(v.title() if i > 0 else v for i, v in enumerate(components[1:]))
        if snake_str[0] == '__':
            components = snake_str.split('_')
            return '__' + ''.join(v.title() if i > 0 else v for i, v in enumerate(components[1:]))
        components = snake_str.split('_')
        return components[0] + ''.join(v.title() for v in components[1:])

    @staticmethod
    def ishex(s: str) -> bool:
        """判断字符串是否为 16 进制字符串。"""
        return all(c in hexdigits for c in s) and len(s) % 2 == 0

    @staticmethod
    def format_hex(hex_str: str, reverse: bool = False) -> str:
        """格式化 16 进制字符串，每 8 位插入一个空格。

        Args:
            hex_str (str): 16 进制字符串.
            reverse (bool, optional): 是否反转字节顺序。 默认为 False.

        Returns:
            str: 格式化后的 16 进制字符串.
        """
        groups = [hex_str[i:i+2] for i in range(0, len(hex_str), 2)]

        if reverse:
            groups.reverse()

        return ' '.join(groups)

    @staticmethod
    def int_to_str(num: int, base: int = 10, length: int = 1) -> str:
        """将整数转换为指定进制的字符串，并在左侧填充 0 到指定长度。

        Args:
            num (int): 整数.
            base (int, optional): 进制. 默认为 10.
            length (int, optional): 填充长度. 默认为 1.

        Returns:
            str: 转换后的字符串.
        """
        result = ''

        if num == 0:
            result = '0' * length
        elif num > 0:
            while num > 0:
                result = StringUtil.base_table[num % base] + result
                num //= base
        else:
            result = '-'
            result += StringUtil.int_to_str(abs(num), base, length - 1)

        if length > 0:
            result = '0' * (length - len(result)) + result

        return result

    @staticmethod
    def str_to_int(string: str, base: int = 10) -> int:
        """将字符串转换为整数，并以指定进制进行计算。

        Args:
            string (str): 字符串.
            base (int, optional): 进制. 默认为 10.

        Returns:
            int: 转换后的整数.
        """
        return int(string, base)

    @staticmethod
    def ip_to_hex(ip_str: str) -> str:
        """将 IP 地址转换为 16 进制字符串。

        Args:
            ip_str (str): IP 地址.

        Returns:
            str: 转换后的 16 进制字符串.
        """
        split_ip = ip_str.split('.')

        if not len(split_ip) == 4 or not all(i.isdigit() and 0 <= int(i) <= 255 for i in split_ip):
            raise ValueError(f'ip_str ({ip_str}) 格式错误')

        return ''.join([StringUtil.int_to_str(int(i), 16, 2) for i in split_ip])

    @staticmethod
    def hex_to_ip(hex_str: str) -> str:
        """将 16 进制字符串转换为 IP 地址。

        Args:
            hex_str (str): 16 进制字符串.

        Returns:
            str: 转换后的 IP 地址.
        """
        if len(hex_str)!= 8:
            raise ValueError(f'hex_str ({hex_str}) 长度必须为 8')
        if not StringUtil.ishex(hex_str):
            raise ValueError(f'hex_str ({hex_str}) 格式错误')

        return '.'.join([str(int(hex_str[i:i+2], 16)) for i in range(0, 8, 2)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, TypeVar, Optional, Literal, Callable, Union, Any
from typing_extensions import Self
from pathlib import Path
from copy import deepcopy

if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath

from ..utils.path_util import PathUtil
from .magic import Magic


class Config(Magic):
    """将配置文件转为 dict 并让其可以像访问对象属性一样访问配置属性"""

    hook = TypeVar('hook', bound=Callable[[Self, Any], Any])

    def __init__(
        self,
        config: Union[dict, 'FileDescriptorOrPath'],
        *,
        decoder: Optional[Literal['yaml', 'toml', 'json']] = None,
        default_map: dict = {},
        hook: Optional['Config.hook'] = None,
        **kw: Any,
    ):
        """
        Args:
            config (Union[dict, 'FileDescriptorOrPath']):
                配置，可以是字典或文件路径.
            decoder (Optional[Literal[&#39;yaml&#39;, &#39;toml&#39;, &#39;json&#39;]], optional):
                当 *config* 参数为文件路径时的解码器. 默认为 None.
            default_map (dict, optional):
                根据 *default_map* 参数设置配置项默认值. 默认为 {}.
            hook (Optional[Callable[[Self, Any], Any]], optional):
                设置属性时的钩子函数, 传入(对象实例, 属性值), 返回
                属性值设置对应属性. 默认为 None.

        Raises:
            ValueError: 参数错误, *config* 必须为字典或路径.
        """
        if decoder and decoder not in {'yaml', 'toml', 'json'}:
            raise ValueError(f'参数错误, 不支持的解码器类型 "{decoder}"')
        if not isinstance(default_map, dict):
            raise ValueError('参数错误, default_map 必须为字典.')
        if hook and not callable(hook):
            raise ValueError('参数错误, hook 必须为可调用对象')

        self.__hook           = hook
        self.__path           = kw.get('path', 'Config')
        self.__name           = kw.get('name', 'Config')
        self.__value          = kw.get('value', None)
        self.__original_value = deepcopy(config)
        self.__file_path      = None

        if config is not None and decoder is not None:
            self.__file_path = PathUtil.abspath(config)
            if not isinstance((_config := self.__load(config, decoder)), dict):
                raise ValueError(f'配置文件 "{self.__file_path}" 无法解析为字典.')
            self.__init_with_dict(_config, default_map)
        elif isinstance(config, dict):
            self.__init_with_dict(config, default_map)
        elif isinstance(config, list):
            if self.__path == 'Config':
                raise ValueError('参数错误, "list" 类型不能作为根配置.')
            self.__init_with_list(config)
        else:
            if self.__path == 'Config':
                raise ValueError(f'参数错误, "{type(config).__name__}" 类型不能作为根配置.')
            self.__init_with_other(config)

    @property
    def path(self) -> str:
        return self.__path

    @property
    def name(self) -> str:
        return self.__name

    @property
    def value(self) -> Any:
        return self.__value

    @property
    def original_value(self) -> Any:
        return self.__original_value

    @property
    def file_path(self) -> Optional[Path]:
        return self.__file_path

    def get_config(self, path: str, default: Any = None) -> Optional['Config']:
        if not str or not isinstance(path, str):
            raise ValueError('参数错误, path 必须为字符串.')

        parts = path.replace('[', '.').replace(']', '').split('.')
        obj = None

        for part in parts:
            if part == 'Config':
                obj = self
                continue

            if not isinstance(obj, Config):
                return default

            if part.isdigit():
                part = int(part)

            if part not in obj:
                return default

            obj = obj[part]

        return obj

    def set_config(self, name: str, value: Any) -> Self:
        self.__dict__[name] = Config(
            value,
            hook=self.__hook,
            path=f'{self.__path}.{name}',
            name=name,
        )
        self.__original_value[name] = value

        return self

    def __getattr__(self, name: str):
        raise AttributeError(f'没有配置属性 "{self.__path}.{name}".')

    def __getattribute__(self, name: str) -> Union['Config', Any]:
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any):
        if not name.startswith('_Config__'):
            raise AttributeError('请使用 Config.set_config 方法设置配置属性.')
        return super().__setattr__(name, value)

    def __getitem__(self, name: str | int) -> Union['Config', Any]:
        if isinstance(name, str) and hasattr(self, name):
            return getattr(self, name)
        return self.__value[name]

    def __setitem__(self, name: str, value: Any) -> None:
        if isinstance(name, str):
            _name = f'.{name}'
        if isinstance(name, int):
            _name = f'[{name}]'
        raise ValueError(f'不允许修改配置项 "{self.__path}{_name}".')

    def __delitem__(self, name: str) -> None:
        del self.__value[name]

    def __len__(self) -> int:
        return len(self.__value)

    def __iter__(self):
        return iter(self.__value)

    def __contains__(self, item: Any):
        if isinstance(self.original_value, dict):
            return hasattr(self, item)
        if isinstance(self.original_value, list):
            return item < len(self.__value)
        return False

    def __load(
        self,
        fp: 'FileDescriptorOrPath',
        decoder: Literal['yaml', 'toml', 'json'],
    ) -> Any:
        match decoder:
            case 'yaml':
                from ..utils.yaml_util import YAMLUtil

                return YAMLUtil.load(fp)
            case 'toml':
                from ..utils.toml_util import TOMLUtil

                return TOMLUtil.load(fp)
            case 'json':
                from ..utils.json_util import JSONUtil

                return JSONUtil.load(fp)

    def __init_with_other(self, config: Any):
        if self.__hook:
            config = self.__hook(self, config)

        if config is ...:
            return

        self.__value = config

    def __init_with_list(self, config: list):
        self.__value = []

        for index, item in enumerate(config):
            self.__value.append(Config(
                item,
                hook=self.__hook,
                path=f'{self.__path}[{index}]',
                name=f'{self.__name}[{index}]',
            ))

    def __init_with_dict(self, config: dict, default_map: dict):
        self.__value = self

        # 设置默认值
        for attr, value in default_map.items():
            self.set_config(attr, value)

        for attr, value in config.items():
            self.set_config(attr, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

class CR4Util:
    """CR4 算法工具类"""

    @staticmethod
    def cr4(key: str | bytes, data: str | bytes, encoding: str = 'utf-8') -> str | bytes:
        """使用 CR4 算法加密 (str) 或解密 (bytes) 数据

        Args:
            key (str | bytes): 密钥
            data (str | bytes): 待加密或解密的数据
            encoding (str, optional): 编码格式. 默认为 'utf-8'.

        Returns:
            str | bytes: 加密 (bytes) 或解密 (str) 后的数据
        """
        _key = bytes(key, encoding) if isinstance(key, str) else key
        _data = bytes(data, encoding) if isinstance(data, str) else data

        S, j = list(range(256)), 0
        for i in range(256):
            j = (j + S[i] + _key[i % len(_key)]) % 256
            S[i], S[j] = S[j], S[i]

        i, j, result = 0, 0, []
        for char in _data:
            i = (i + 1) % 256
            j = (j + S[i]) % 256
            S[i], S[j] = S[j], S[i]
            result.append(char ^ S[(S[i] + S[j]) % 256])

        return bytes(result) if isinstance(data, str) else bytes(result).decode(encoding)

    @staticmethod
    def encrypt(key: str | bytes, data: str, encoding: str = 'utf-8') -> bytes:
        """加密数据

        Args:
            key (str | bytes): 密钥
            data (str): 待加密的数据
            encoding (str, optional): 编码格式. 默认为 'utf-8'.

        Returns:
            bytes: 加密后的数据
        """
        return CR4Util.cr4(key, data, encoding)

    @staticmethod
    def decrypt(key: str | bytes, data: bytes, encoding: str = 'utf-8') -> str:
        """解密数据

        Args:
            key (str | bytes): 密钥
            data (bytes): 待解密的数据
            encoding (str, optional): 编码格式. 默认为 'utf-8'.

        Returns:
            str: 解密后的数据
        """
        return CR4Util.cr4(key, data, encoding)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from datetime import timedelta, datetime, timezone


class DateTimeUtil(object):
    """日期时间工具类"""

    timedelta = timedelta
    """等价于 datetime.timedelta"""

    @staticmethod
    def now() -> datetime:
        """返回 datetime 格式的当前时间"""
        return datetime.now(tz=timezone.utc)

    @staticmethod
    def str_now(format=r'%Y-%m-%d %H:%M:%S') -> str:
        """返回字符串格式的当前时间"""
        return datetime.now(tz=timezone.utc).strftime(format)

    @staticmethod
    def diff(minuend: str, minus: str, format=r'%Y-%m-%d %H:%M:%S') -> timedelta:
        """返回 minuend 减去 minus 的时间差 (单位 s)"""
        return (datetime.strptime(minuend, format) - datetime.strptime(minus, format))

    @staticmethod
    def strftime(time: datetime, format=r'%Y-%m-%d %H:%M:%S') -> str:
        """datetime 时间格式转字符串"""
        return datetime.strftime(time, format)

    @staticmethod
    def strptime(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> datetime:
        """字符串时间格式转 datetime"""
        return datetime.strptime(stime, format)

    @staticmethod
    def stime2year(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> int:
        """返回字符串时间格式中的 年"""
        return datetime.strptime(stime, format).year

    @staticmethod
    def stime2month(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> int:
        """返回字符串时间格式中的 月"""
        return datetime.strptime(stime, format).month

    @staticmethod
    def stime2day(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> int:
        """返回字符串时间格式中的 日"""
        return datetime.strptime(stime, format).day

    @staticmethod
    def stime2hour(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> int:
        """返回字符串时间格式中的 时"""
        return datetime.strptime(stime, format).hour

    @staticmethod
    def stime2minute(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> int:
        """返回字符串时间格式中的 分"""
        return datetime.strptime(stime, format).minute

    @staticmethod
    def stime2second(stime: str, format=r'%Y-%m-%d %H:%M:%S') -> int:
        """返回字符串时间格式中的 秒"""
        return datetime.strptime(stime, format).second



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Mapping
from re import Match, sub
import logging

from .color import AnsiColor


class ColorFormatter(logging.Formatter):
    """带颜色的日志格式化类"""

    COLOR  = {
        logging.DEBUG: AnsiColor.RESET_ALL,
        logging.INFO: AnsiColor.FORE_GREEN,
        logging.WARNING: AnsiColor.FORE_YELLOW,
        logging.ERROR: AnsiColor.FORE_RED,
        logging.CRITICAL: AnsiColor.FORE_MAGENTA,
    }

    def __init__(self, *args, **kwargs) -> None:
        # 修改格式化 levelname 的长度
        if 'fmt' in kwargs:
            def replace(match: Match) -> str:
                _g1 = match.group(1)
                _g2 = int(match.group(2))
                return f'%(levelname){_g1}{_g2 + 10}s'
            kwargs['fmt'] = sub(r'%\(levelname\)(\S*)(\d)s', replace, kwargs['fmt'])

        super().__init__(*args, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        color = self.COLOR[record.levelno]
        record.levelname = f'{color}{record.levelname}{AnsiColor.RESET_ALL}'
        return super().format(record)


class Logger:
    """日志类

    Args:
        name (str, optional): 日志名称. 默认为 'easy_pyoc'.
        formatter (str, optional): 日志格式. 默认为 '[%(asctime)s] %(levelname)-8s : %(message)s'.
        use_color (bool, optional): 是否使用颜色. 默认为 True.

    Returns:
        logging.Logger: 日志实例
    """
    _instance: Mapping[str, logging.Logger] = {}

    def __new__(cls, *args, **kwargs):
        name      = kwargs.get('name', 'easy_pyoc')
        fmt       = kwargs.get('fmt', '[%(asctime)s] %(levelname)-8s : %(message)s')
        use_color = kwargs.get('use_color', True)

        if name not in cls._instance:
            cls._instance[name] = logging.getLogger(name)
            cls._instance[name].setLevel(logging.INFO)

            if fmt is not None:
                handler   = logging.StreamHandler()
                formatter = ColorFormatter(fmt=fmt) if use_color else logging.Formatter(fmt)
                formatter.default_msec_format = '%s.%03d'
                handler.setFormatter(formatter)
                cls._instance[name].addHandler(handler)

        return cls._instance[name]


if __name__ == '__main__':
    # color 测试，运行前请修改导入
    for use_color in [True, False]:
        logger = Logger(name=f'easy_pyoc_{use_color}', fmt='%(levelname)-8s : %(message)s', use_color=use_color)
        logger.setLevel(logging.DEBUG)

        logger.debug('hello, world')
        logger.info('hello, world')
        logger.warning('hello, world')
        logger.error('hello, world')
        logger.critical('hello, world')
        print('-' * 80)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest, re

from easy_pyoc import KNXUtil
from easy_pyoc import NetworkUtil
from easy_pyoc import StringUtil
from easy_pyoc import ThreadUtil
from easy_pyoc import Config
from easy_pyoc import ExceptionUtil
from easy_pyoc import CR4Util


def test_knx_util():
    for device in KNXUtil.discover(2):
        assert isinstance(device, KNXUtil.ControlDevice)
        assert device.host
        assert device.port


def test_network_util():
    assert NetworkUtil.is_ip('192.168.1.1')
    assert NetworkUtil.is_ip('255.255.255.255')
    assert NetworkUtil.is_ip('0.0.0.0')
    assert not NetworkUtil.is_ip('192.168.1')
    assert not NetworkUtil.is_ip('192.168.1.1.1')

    assert NetworkUtil.is_host('www.baidu.com')
    assert NetworkUtil.is_host('192.168.1.1')
    assert not NetworkUtil.is_host('www.baidu')

    assert NetworkUtil.classify_ip('10.0.0.1')     == 'A'
    assert NetworkUtil.classify_ip('172.16.0.1')   == 'B'
    assert NetworkUtil.classify_ip('192.168.1.1')  == 'C'
    assert NetworkUtil.classify_ip('224.168.3.11') == 'D'
    assert NetworkUtil.classify_ip('240.168.3.11') == 'E'
    with pytest.raises(ValueError, match='无效的 IP 地址'):
        NetworkUtil.classify_ip('192.168.1')

    assert NetworkUtil.get_local_ip()
    with pytest.raises(ValueError, match='无效的网卡序号'):
        NetworkUtil.get_local_ip(99)

    assert NetworkUtil.get_broadcast_address('192.168.1.0/24') == '192.168.1.255'
    assert NetworkUtil.get_broadcast_address('192.168.1.1/32') == '192.168.1.1'
    with pytest.raises(ValueError):
        NetworkUtil.get_broadcast_address('192.168.1')
    with pytest.raises(ValueError):
        NetworkUtil.get_broadcast_address('192.168.1.0/33')


def test_string_util():
    assert StringUtil.int_to_str(123, 2) == '1111011'
    assert StringUtil.int_to_str(123, 2, length=5) == '1111011'
    assert StringUtil.int_to_str(123, 2, length=8) == '01111011'
    assert StringUtil.int_to_str(0, 2) == '0'
    assert StringUtil.int_to_str(0, 2, length=5) == '00000'
    assert StringUtil.int_to_str(-123, 2) == '-1111011'
    assert StringUtil.int_to_str(123456789, 2) == '111010110111100110100010101'

    assert StringUtil.int_to_str(123, 8) == '173'
    assert StringUtil.int_to_str(123, 8, length=5) == '00173'
    assert StringUtil.int_to_str(0, 8) == '0'
    assert StringUtil.int_to_str(0, 8, length=5) == '00000'
    assert StringUtil.int_to_str(-123, 8) == '-173'
    assert StringUtil.int_to_str(123456789, 8) == '726746425'

    assert StringUtil.int_to_str(123) == '123'
    assert StringUtil.int_to_str(123, length=5) == '00123'
    assert StringUtil.int_to_str(0) == '0'
    assert StringUtil.int_to_str(0, length=5) == '00000'
    assert StringUtil.int_to_str(-123) == '-123'
    assert StringUtil.int_to_str(123456789) == '123456789'

    assert StringUtil.int_to_str(123, base=16) == '7b'
    assert StringUtil.int_to_str(123, base=16, length=5) == '0007b'
    assert StringUtil.int_to_str(0, base=16) == '0'
    assert StringUtil.int_to_str(0, base=16, length=5) == '00000'
    assert StringUtil.int_to_str(123456789, base=16) == '75bcd15'
    assert StringUtil.int_to_str(123456789, base=10) == '123456789'
    assert StringUtil.int_to_str(123456789, base=16, length=8) == '075bcd15'

    assert StringUtil.str_to_int('1111011', 2) == 123
    assert StringUtil.str_to_int('01111011', 2) == 123
    assert StringUtil.str_to_int('-1111011', 2) == -123

    assert StringUtil.format_hex('7b') == '7b'
    assert StringUtil.format_hex('7bcd15') == '7b cd 15'

    assert StringUtil.ip_to_hex('192.168.1.1') == 'c0a80101'
    assert StringUtil.ip_to_hex('255.255.255.255') == 'ffffffff'
    assert StringUtil.ip_to_hex('0.0.0.0') == '00000000'

    assert StringUtil.hex_to_ip('c0a80101') == '192.168.1.1'
    assert StringUtil.hex_to_ip('ffffffff') == '255.255.255.255'
    assert StringUtil.hex_to_ip('00000000') == '0.0.0.0'


def test_thread_util():
    from time import sleep

    n = 0
    def task(stop_flag):
        nonlocal n

        while stop_flag.is_set():
            print('running...')
            n += 1
            sleep(1)

    task_id = ThreadUtil.execute_task(task)
    sleep(3)
    ThreadUtil.chancel_task(task_id)
    assert n == 3


def test_toml_util():
    import sys

    from easy_pyoc import TOMLUtil

    if sys.version_info < (3, 11):
        with pytest.raises(AttributeError):
            TOMLUtil.loads('')

        with pytest.raises(ImportError, match='没有成功导入'):
            _ = TOMLUtil()
    else:
        assert TOMLUtil.loads('') == {}


def test_config_class():
    with pytest.raises((ImportError, FileNotFoundError), match='(无法导入|文件.+不存在)'):
        Config('test.toml', decoder='toml')

    with pytest.raises(FileNotFoundError, match='文件 .+ 不存在'):
        Config('test.json', decoder='json')

    with pytest.raises(ValueError, match='不支持的解码器类型 .+'):
        Config('test.txt', decoder='txt')

    with pytest.raises(ValueError, match='不能作为根配置'):
        Config([])

    with pytest.raises(ValueError, match='不能作为根配置'):
        Config(1)

    def hook(o: Config, v):
        if isinstance(v, int) and o.path != 'Config.f':
            return v * 2
        if o.path == 'Config.b[1][1].a1':
            return 'aaa'
        if o.path == 'Config.f':
            return 666
        return v

    _ = Config(
        {'a': 333, 'b': [1, [2, {'a1': 'a1'}]], 'c': {'c1': 'c1', 'c2': ['c21', 'c22']}},
        default_map={'a': 123, 'f': 456},
    ).set_config('d', 1).set_config('e', 2)
    # 根测试
    assert _.path == 'Config'
    assert _.name == 'Config'
    assert _.value == _
    assert _.original_value == {'a': 333, 'b': [1, [2, {'a1': 'a1'}]], 'c': {'c1': 'c1', 'c2': ['c21', 'c22']}, 'f': 456, 'd': 1, 'e': 2}
    # 简单子测试
    assert _.a.path == 'Config.a'
    assert _.a.name == 'a'
    assert _.a.value == 333
    assert _.a.original_value == 333
    # 列表测试
    assert _.b.path == 'Config.b'
    assert _.b.name == 'b'
    assert _.b.value == [_.b[0], _.b[1]]
    assert _.b.original_value == [1, [2, {'a1': 'a1'}]]
    # 列表子测试
    assert _.b[0].path == 'Config.b[0]'
    assert _.b[0].name == 'b[0]'
    assert _.b[0].value == 1
    assert _.b[0].original_value == 1
    # 字典测试
    assert _.c.path == 'Config.c'
    assert _.c.name == 'c'
    assert _.c.value == _.c
    assert _.c.original_value
    # 字符串取值测试
    assert _['c'].path == 'Config.c'
    assert _['c'].name == 'c'
    assert _['c'].value == _.c
    assert _['c'].original_value
    # 字典子测试
    assert _.c.c1.path == 'Config.c.c1'
    assert _.c.c1.name == 'c1'
    assert _.c.c1.value == 'c1'
    assert _.c.c1.original_value == 'c1'
    # 嵌套列表测试
    assert _.b[1][1].path == 'Config.b[1][1]'
    assert _.b[1][1].name == 'b[1][1]'
    assert _.b[1][1].value == _.b[1][1]
    assert _.b[1][1].original_value == {'a1': 'a1'}
    # 嵌套字典测试
    assert _.b[1][1].a1.path == 'Config.b[1][1].a1'
    assert _.b[1][1].a1.name == 'a1'
    assert _.b[1][1].a1.value == 'a1'
    assert _.b[1][1].a1.original_value == 'a1'
    # 自定义属性测试
    assert _.d.path == 'Config.d'
    assert _.d.name == 'd'
    assert _.d.value == 1
    assert _.d.original_value == 1
    assert _.e.path == 'Config.e'
    assert _.e.name == 'e'
    assert _.e.value == 2
    assert _.e.original_value == 2
    # 默认值测试
    assert _.a.value == 333
    assert _.f.value == 456
    # 递归取值测试
    assert _.get_config('Config.a').value == 333
    assert _.get_config('Config.b[0]').value == 1
    assert _.get_config('Config.b[1][0]').value == 2
    assert _.get_config('Config.b[1][1].a1').value == 'a1'
    assert _.get_config('Config.c.c1').value == 'c1'
    # hook 测试
    _ = Config(
        {'a': 333, 'b': [1, [2, {'a1': 'a1'}]], 'c': {'c1': 'c1', 'c2': ['c21', 'c22']}},
        default_map={'a': 123, 'f': 456},
        hook=hook,
    ).set_config('d', 1).set_config('e', 2)
    assert _.a.value == 666
    assert _.a.original_value == 333
    assert _.b[1][1].a1.value == 'aaa'
    assert _.b[1][1].a1.original_value == 'a1'
    assert _.f.value == 666
    # 未知属性测试
    with pytest.raises(AttributeError, match=re.escape('没有配置属性 "Config.x".')):
        _.x
    with pytest.raises(AttributeError, match=re.escape('没有配置属性 "Config.b[1][1].a2".')):
        _.b[1][1].a2
    # 字典属性修改测试
    with pytest.raises(AttributeError, match=re.escape('请使用 Config.set_config 方法设置配置属性.')):
        _.t1 = 123
    with pytest.raises(ValueError, match=re.escape('不允许修改配置项 "Config.c.c1".')):
        _.c['c1'] = 12345
    with pytest.raises(ValueError, match=re.escape('不允许修改配置项 "Config.b[0]".')):
        _.b[0] = 12345


def test_exception_util():
    try:
        raise ValueError('值错误')
    except:
        assert ExceptionUtil.get_message() == 'ValueError: 值错误'


def test_cr4_util():
    key = 'secret_key'
    d_data = 'hello, world!'
    e_data = b'+Z\xf7\xcd\xafy\xcbQ&\x92\xd8\x0f\xe6'

    assert CR4Util.cr4(key, d_data) == e_data
    assert CR4Util.cr4(key, e_data) == d_data
    assert CR4Util.encrypt(key, d_data) == e_data
    assert CR4Util.decrypt(key, e_data) == d_data

    assert CR4Util.cr4(bytes(key, 'utf-8'), d_data) == e_data
    assert CR4Util.cr4(bytes(key, 'utf-8'), e_data) == d_data
    assert CR4Util.encrypt(bytes(key, 'utf-8'), d_data) == e_data
    assert CR4Util.decrypt(bytes(key, 'utf-8'), e_data) == d_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

from easy_pyoc import XMLUtil


X1 = '''
<item />
<item />
<item />
'''
D1 = {
    "item": [
        "",
        "",
        "",
    ]
}

X2 = '''
<root />
'''
D2 = {
    "root": None
}

X21 = '''
<root id="1" />
'''
D21 = {
    "root": {
        "@id": 1
    }
}

X22 = '''
<root></root>
'''
X221 = '''
<root />
'''
D22 = {
    "root": None
}

X23 = '''
<root>
</root>
'''
X231 = '''
<root />
'''
D23 = {
    "root": None
}

X3 = '''
<root>
    <item />
</root>
'''
D3 = {
    "root": {
        "item": None
    }
}

X31 = '''
<root>
    <item />
    <item />
    <item />
</root>
'''
D31 = {
    "root": {
        "item": [
            None,
            None,
            None
        ]
    }
}

X32 = '''
<root>
    <item />
    <item>
    </item>
    <item id="12" />
    <item>123</item>
</root>
'''
X321 = '''
<root><item/><item/><itemid="12"/><item>123</item></root>
'''
D32 = {
    "root": {
        "item": [
            None,
            None,
            {
                "@id": 12,
            },
            123
        ]
    }
}

X5 = '''
<root>
    <item id="1">
        <name>item1</name>
        <value>100.1</value>
        <readonly>true</readonly>
    </item>
    <item id="2">
        <name>item2</name>
        <value>200.2</value>
        <readonly>false</readonly>
    </item>
</root>
'''
X51 = '''
<root>
    <item id="1">
        <name>item1</name>
        <value>100.1</value>
        <readonly>True</readonly>
    </item>
    <item id="2">
        <name>item2</name>
        <value>200.2</value>
        <readonly>False</readonly>
    </item>
</root>
'''
D5 = {
    "root": {
        "item": [
            {
                "@id": 1,
                "name": "item1",
                "value": 100.1,
                "readonly": True
            },
            {
                "@id": 2,
                "name": "item2",
                "value": 200.2,
                "readonly": False
            }
        ]
    }
}

X6 = '''
<items>
    <item id="1">
        _id_1_text
        <name>item1</name>
        <value>100</value>
        <subs>
            <sub id="11">
                _id_11_text
                <name>sub11</name>
                <value>110</value>
            </sub>
            <sub id="12">
                <name>sub12</name>
                <value>120</value>
                id_12_text
            </sub>
        </subs>
        id_1_text
    </item>
    <item id="2">

        _id_2_text
        <name>item2</name>
        <value>200</value>
        <subs>
            <sub id="21">
                21
                <name>sub21</name>
                <value>210</value>
                211
            </sub>
            <sub id="22">
                22
                <name>sub22</name>
                222
                <value>220</value>
                223
            </sub>
        </subs>
        id_2_text

    </item>
</items>
'''
X61 = '''
<items>
    <item id="1">
        _id_1_text
        id_1_text
        <name>item1</name>
        <value>100</value>
        <subs>
            <sub id="11">
                _id_11_text
                <name>sub11</name>
                <value>110</value>
            </sub>
            <sub id="12">
                id_12_text
                <name>sub12</name>
                <value>120</value>
            </sub>
        </subs>
    </item>
    <item id="2">

        _id_2_text
        id_2_text
        <name>item2</name>
        <value>200</value>
        <subs>
            <sub id="21">
                21
                211
                <name>sub21</name>
                <value>210</value>
            </sub>
            <sub id="22">
                22
                222
                223
                <name>sub22</name>
                <value>220</value>
            </sub>
        </subs>

    </item>
</items>
'''
D6 = {
    "items": {
        "item": [
            {
                "@id": 1,
                "name": "item1",
                "value": 100,
                "subs": {
                    "sub": [
                        {
                            "@id": 11,
                            "name": "sub11",
                            "value": 110,
                            "#text": "_id_11_text"
                        },
                        {
                            "@id": 12,
                            "name": "sub12",
                            "value": 120,
                            "#text": "id_12_text"
                        }
                    ]
                },
                "#text": "_id_1_text\n        \n        \n        \n        id_1_text"
            },
            {
                "@id": 2,
                "name": "item2",
                "value": 200,
                "subs": {
                    "sub": [
                        {
                            "@id": 21,
                            "name": "sub21",
                            "value": 210,
                            "#text": "21\n                \n                \n                211"
                        },
                        {
                            "@id": 22,
                            "name": "sub22",
                            "value": 220,
                            "#text": "22\n                \n                222\n                \n                223"
                        }
                    ]
                },
                "#text": "_id_2_text\n        \n        \n        \n        id_2_text"
            }
        ]
    }
}


def test_xml_to_dict():
    with pytest.raises(ValueError, match='格式有误'):
        XMLUtil.xml_to_dict(X1)
    assert XMLUtil.xml_to_dict(X2) == D2
    assert XMLUtil.xml_to_dict(X21) == D21
    assert XMLUtil.xml_to_dict(X22) == D22
    assert XMLUtil.xml_to_dict(X23) == D23
    assert XMLUtil.xml_to_dict(X3) == D3
    assert XMLUtil.xml_to_dict(X31) == D31
    assert XMLUtil.xml_to_dict(X32) == D32
    assert XMLUtil.xml_to_dict(X5) == D5
    assert XMLUtil.xml_to_dict(X6) == D6


def test_dict_to_xml():
    import re
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D1)) == re.sub(r'\s', '', X1)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D2)) == re.sub(r'\s', '', X2)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D21)) == re.sub(r'\s', '', X21)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D22)) == re.sub(r'\s', '', X221)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D23)) == re.sub(r'\s', '', X231)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D3)) == re.sub(r'\s', '', X3)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D31)) == re.sub(r'\s', '', X31)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D32)) == re.sub(r'\s', '', X321)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D5)) == re.sub(r'\s', '', X51)
    assert re.sub(r'\s', '', XMLUtil.dict_to_xml(D6)) == re.sub(r'\s', '', X61)


if __name__ == '__main__':
    pytest.main(['-s', __file__])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TypeVar, Literal
from functools import reduce
from operator import getitem

from .string_util import StringUtil


_T = TypeVar('_T')


class ObjectUtil(object):
    """对象工具类"""

    @staticmethod
    def repr(obj: object, exclude: set[str] = {}, include: set[str] = {}) -> str:
        """将对象转为描述属性的字符串

        Args:
            obj (object): _description_
            exclude (set[str], optional): 忽略属性列表. 默认为 {}.
            include (set[str], optional): 包含属性列表. 默认为 {}.

        Returns:
            str: 对象详情字符串
        """
        class_name = obj.__class__.__name__
        attributes = [
            f'{k}={f"{v}" if isinstance(v, str) else v}'
            for k, v in obj.__dict__.items()
            if k not in exclude
                and not k.startswith('_')
                or k in include
        ]
        attributes_str = ', '.join(attributes)
        return f'{class_name}({attributes_str})'

    @staticmethod
    def vars(obj: object, exclude: set[str] = {}, include: set[str] = {}, style='snake') -> dict:
        """将对象的属性转为字典形式

        Args:
            obj (object): 对象实例
            exclude (set[str], optional): 忽略属性列表. 默认为 {}.
            include (set[str], optional): 包含属性列表. 默认为 {}.
            style (str, optional): 键名命名风格，为 None 不转换. 默认为 'snake'.

        Returns:
            dict: 对象 { 属性: 值 } 组成的字典
        """
        dict_items = {}
        if isinstance(obj, dict):
            dict_items = obj.items()
        if hasattr(obj, '__dict__'):
            dict_items = obj.__dict__.items()

        if style == 'snake':
            return {
                StringUtil.camel_to_snake(k): v
                for k, v in dict_items
                if k not in exclude
                    and not k.startswith('_')
                    or k in include
            }
        if style == 'camel':
            return {
                StringUtil.snake_to_camel(k): v
                for k, v in dict_items
                if k not in exclude
                    and not k.startswith('_')
                    or k in include
            }
        return {
            k: v
            for k, v in dict_items
            if k not in exclude
                and not k.startswith('_')
                or k in include
        }

    @staticmethod
    def update_with_dict(obj: _T, *, exclude: set[str] = {}, style: Literal['snake', 'camel'] = 'snake', **kw) -> _T:
        """用关键字参数将对象赋值, 并忽略对象上不存在的属性

        Args:
            obj (_T): 对象实例
            exclude (set[str], optional): 忽略属性列表. 默认为 {}.
            style (Literal[&#39;snake&#39;, &#39;camel&#39;], optional): 键名命名风格. 默认为 'snake'.
            **kw: 关键字参数

        Returns:
            _T: 更新数据后的对象实例
        """
        style = kw.get('style', 'snake')
        exclude = kw.get('exclude', [])
        for k, v in kw.items():
            if style == 'snake':
                k = StringUtil.camel_to_snake(k)
            if style == 'camel':
                k = StringUtil.snake_to_camel(k)
            if hasattr(obj, k) and k not in exclude:
                setattr(obj, k, v)
        return obj

    @staticmethod
    def get_value_from_dict(data_dict: dict, key_string: str, default = None) -> any:
        """根据带点字符串的层级取出字典中的数据

        Args:
            data_dict (dict): 类字典类型
            key_string (str): 以英文句号分隔的字符串
            default (_type_, optional): 默认值. 默认为 None.

        Returns:
            any: 字典中对应的值
        """
        keys = key_string.split('.')
        try:
            return reduce(getitem, keys, data_dict)
        except:
            return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class NotThisModule(object):

    def __new__(cls, *args, **kwargs):
        raise ImportError(cls.get_msg())

    def __repr__(self):
        raise ImportError(self.get_msg())

    def __getattribute__(self, name):
        raise ImportError(self.get_msg())

    @classmethod
    def get_msg(cls):
        return f'没有成功导入 "{cls.get_name()}", 原因: "{cls.get_reason()}".'

    @classmethod
    def get_name(cls):
        return cls.__not_this_module__['name']

    @classmethod
    def get_reason(cls):
        return cls.__not_this_module__['reason']


def not_this_module(name: str, reason: str):
    return type('NotThisModule', (NotThisModule,), {
        '__not_this_module__': {'name': name, 'reason': reason},
    })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""基于 PyQt5 的 GUI.


classes:
    class OTitleBar(QWidget)
    自定义标题栏，不可单独使用.

    class OWindow(QMainWindow)
    自定义窗口，内置 自定义标题栏及 qss;
    1. 实现了自定义标题栏的主窗口，包含以下功能：
        a. 窗口最大化、最小化、关闭;
        b. 窗口移动;
        c. 拖动改变窗口大小;
    2. 基于 requests.get 实现了从互联网及本地加载背景图片.
"""

from PyQt5.QtCore import Qt, QObject, QEvent, QSize, QRect
from PyQt5.QtGui import QIcon, QPixmap, QPalette, QBrush
from PyQt5.QtWidgets import *
from requests import get
from os.path import exists
import logging

from . import OcGuiRes


__auth__ = 'one-ccs'

logging.basicConfig(format='%(levelname)s: %(message)s')


class OTitleBar(QWidget):
    """自定义标题栏

    实现了最大化与最小化"""

    def __init__(self, parent):
        super().__init__()
        # 设置尺寸策略
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.setFixedHeight(36)
        # ~ self.setAttribute(Qt.WA_TransparentForMouseEvents)  # 鼠标点击穿透

        self._parent = parent                 # 父（主）窗口
        self._menu_button = None              # 菜单按钮（QWidget）
        self._skin_button = None              # 换肤按钮（QWidget）
        self._title_icon = None               # 图标（QLabel）
        self.title_label = None               # 标题（QLabel）
        self._min_button = None               # 最小化按钮（QWidget）
        self._max_button = None               # 最大化按钮（QWidget）
        self._close_button = None             # 关闭按钮（QWidget）
        self._central_layout = None           # 主布局（水平）

        self._menu_icon = None                # 菜单图标
        self._skin_icon = None                # 换肤图标
        self._min_icon = None                 # 最小化按钮图标
        self._max_icon = None                 # 最大化按钮图标
        self._restore_icon = None             # 最大化按钮恢复图标
        self._close_icon = None               # 关闭按钮图标

        self._series = 'black'                # 图标样式（默认黑色）

        # 初始化标题栏属性
        self.init_titlebar()
        # 设置最窄宽度
        self.setMinimumWidth(self._title_icon.width() * 13)
        # 去除布局及控件间的间隔及边距
        set_all_gap(self._central_layout, left=8)
        self.setMouseTracking(True)
        self._title_icon.setMouseTracking(True)
        self.title_label.setMouseTracking(True)

    def init_titlebar(self):
        """初始化标题栏属性"""
        # 添加标题控件
        self._title_icon = QLabel()
        self.title_label = QLabel()
        self._min_button = QPushButton()
        self._max_button = QPushButton()
        self._close_button = QPushButton()
        self._title_icon.setObjectName('OTitleIcon')
        self.title_label.setObjectName('OTitleLabel')
        self._min_button.setObjectName('OMinButton')
        self._max_button.setObjectName('OMaxButton')
        self._close_button.setObjectName('OCloseButton')

        # 设置 icon
        set_icon(self._title_icon, ':/res/images/one_ccs.ico')
        self._load_icon()

        # 设置 icon 宽度
        self._title_icon.setFixedHeight(24)
        self._title_icon.setFixedWidth(24)

        # 设置标题控件策略
        expanding = QSizePolicy.Expanding
        self.title_label.setSizePolicy(expanding, expanding)

        # 设置窗口主布局，并放入主容器中
        self._central_layout = QHBoxLayout(self)

        # 把控件添加进标题栏布局
        self._central_layout.addWidget(self._title_icon, alignment=Qt.AlignVCenter)
        self._central_layout.addWidget(self.title_label, alignment=Qt.AlignVCenter)
        self._central_layout.addWidget(self._min_button, alignment=Qt.AlignVCenter)
        self._central_layout.addWidget(self._max_button, alignment=Qt.AlignVCenter)
        self._central_layout.addWidget(self._close_button, alignment=Qt.AlignVCenter)

        # 链接信号/槽
        self._min_button.clicked.connect(self.min_button_click)
        self._max_button.clicked.connect(self.max_button_click)
        self._close_button.clicked.connect(self.close_button_click)

    def set_menu_icon(self, icon):
        """设置菜单按钮图标"""
        if not isinstance(icon, QPixmap):
            icon = QPixmap(icon)

        icon = icon.scaled(width, height, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        self._menu_icon = QIcon(QPixmap(icon))
        self._menu_button.setPixmap()
        self._menu_button.setIcon(self._menu_icon)

    def set_skin_icon(self, icon):
        """设置最换肤按钮图标"""
        self._skin_icon = QIcon(QPixmap(icon))
        self._skin_button.setIcon(self._skin_icon)

    def set_min_icon(self, icon):
        """设置最小化按钮图标"""
        self._min_icon = QIcon(QPixmap(icon))
        self._min_button.setIcon(self._min_icon)

    def set_max_icon(self, icon):
        """设置最大化按钮图标"""
        self._max_icon = QIcon(QPixmap(icon))
        self._max_button.setIcon(self._max_icon)

    def set_restore_icon(self, icon):
        """设置恢复按钮图标"""
        self._restore_icon = QIcon(QPixmap(icon))

    def set_close_icon(self, icon):
        """设置关闭按钮图标"""
        self._close_icon = QIcon(QPixmap(icon))
        self._close_button.setIcon(self._close_icon)

    def set_icon_series(self, series):
        """设置标题图标系列（黑/白）"""
        if series == 'black' or series == 'white':
            self._series = series
            self._load_icon()

    def add_menu_button(self):
        """添加菜单按钮"""
        self._menu_button = QPushButton()
        self.setObjectName('OMenuButton')
        self._menu_button.installEventFilter(self._parent)
        if self._series == 'black':
            self.set_menu_icon(':/res/images/title_menu.ico')
        elif self._series == 'white':
            self.set_menu_icon(':/res/images/title_menu_w.ico')
        self._central_layout.addWidget(self._menu_button)
        # 重新放置主按钮
        self._central_layout.addWidget(self._min_button)
        self._central_layout.addWidget(self._max_button)
        self._central_layout.addWidget(self._close_button)

        return self._menu_button

    def add_skin_button(self):
        """添加换肤按钮"""
        self._skin_button = QPushButton()
        self._skin_button.setObjectName('OSkinButton')
        self._skin_button.installEventFilter(self._parent)
        if self._series == 'black':
            self.set_skin_icon(':/res/images/title_skin.ico')
        elif self._series == 'white':
            self.set_skin_icon(':/res/images/title_skin_w.ico')
        # 设置图标大小
        # ~ self._skin_button.setIconSize(QSize(22, 22))
        self._central_layout.addWidget(self._skin_button)
        # 重新放置主按钮
        self._central_layout.addWidget(self._min_button)
        self._central_layout.addWidget(self._max_button)
        self._central_layout.addWidget(self._close_button)

        return self._skin_button

    def _load_icon(self):
        """设置图标"""

        res = ':/res/images/'
        if self._series == 'black':
            if self._menu_button:
                self.set_menu_icon(res + 'title_menu.ico')
            if self._skin_button:
                self.set_skin_icon(res + 'title_skin.ico')
            self.set_min_icon(res + 'title_min.ico')
            self.set_max_icon(res + 'title_max.ico')
            self.set_restore_icon(res + 'title_restore.ico')
            self.set_close_icon(res + 'title_close.ico')
        elif self._series == 'white':
            if self._menu_button:
                self.set_menu_icon(res + 'title_menu_w.ico')
            if self._skin_button:
                self.set_skin_icon(res + 'title_skin_w.ico')
            self.set_min_icon(res + 'title_min_w.ico')
            self.set_max_icon(res + 'title_max_w.ico')
            self.set_restore_icon(res + 'title_restore_w.ico')
            self.set_close_icon(res + 'title_close_w.ico')

    def _check_parent(self):
        """判断是否有父窗口"""
        if isinstance(self._parent, QWidget) or (
           isinstance(self._parent, QMainWindow)):
            return True
        else:
            return False

    # 槽函数
    def min_button_click(self):
        """最小化"""
        if self._check_parent():
            self._parent.showMinimized()
        else:
            self.showMinimized()

    def max_button_click(self):
        """最大化（恢复）"""
        if self._check_parent():
            if self._parent.isMaximized():
                self._parent.showNormal()
                self._max_button.setIcon(self._max_icon)
            else:
                self._parent.showMaximized()
                self._max_button.setIcon(self._restore_icon)

    def close_button_click(self):
        """关闭"""
        if self._check_parent():
            self._parent.close()
        else:
            self.close()


class OWindow(QMainWindow):
    """这是自定义的窗口"""

    def __init__(self, title='easy_pyoc'):
        super().__init__()
        self.setWindowFlags(Qt.FramelessWindowHint)         # 设置为无标题栏
        # ~ self.setAttribute(Qt.WA_TranslucentBackground)  # 设置窗口背景透明
        # ~ self.setWindowOpacity(0.85)                     # 设置窗口不透明度

        self._super = None                    # 顶级控件（放置背景图片）
        self._central = None                  # 主部件（放置标题栏、工具栏、客户区）
        self._super_layout = None             # 顶级布局（垂直）
        self._central_layout = None           # 中心布局（垂直）

        self.title = title                    # 标题
        self.titlebar = None                  # 标题栏（OTitleBar）
        self.toolbar = None                   # 工具栏（QStackedWidget）
        self.client = None                    # 客户区（QStackedWidget）

        self.MOVE_SENSITIVITY = 3             # 窗口移动反应灵敏度，反比
        self.MARGIN = 9                       # 边缘宽度(拖动改变窗口大小)

        self._bgimg = None                    # 存放背景图片的变量
        self._isPressed = False
        self._press_button = None
        self._last_geometry = None
        self._area = None
        self._move_count = 0

        # 初始化窗口属性
        self.init_window()
        self.setMinimumHeight(self.titlebar.height())
        self.setMinimumWidth(self.titlebar.minimumWidth())
        # 设置布局间隙
        set_all_gap(self._super_layout)
        set_all_gap(self._central_layout)
        # 设置尺寸策略
        self.toolbar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        # 安装事件过滤器
        self.titlebar._min_button.installEventFilter(self)
        self.titlebar._max_button.installEventFilter(self)
        self.titlebar._close_button.installEventFilter(self)
        # 设置鼠标追踪
        self.setMouseTracking(True)
        self._super.setMouseTracking(True)
        self._central.setMouseTracking(True)
        self.toolbar.setMouseTracking(True)
        self.client.setMouseTracking(True)
        # 加载默认样式表
        self.setStyleSheet('')
        # 显示窗口
        self.show()

    def init_window(self):
        """初始化窗口"""

        # 超·部件（不显示）
        self._super = QWidget()
        self._super.setObjectName('OSuper')
        # 重设窗口主控件
        self.setCentralWidget(self._super)

        # 中心控件(显示部分)
        self._central = QWidget()
        self._central.setObjectName('OCentral')
        self._central_layout = QVBoxLayout(self._central)

        # 标题栏
        self.titlebar = OTitleBar(self)
        # 设置标题
        self.setWindowTitle(self.title)

        # 工具栏
        self.toolbar = QStackedWidget()

        # 客户区，放入客户区布局
        self.client = QStackedWidget()
        self.client.setObjectName('OClient')

        # 中心布局添加部件
        self._central_layout.addWidget(self.titlebar)
        self._central_layout.addWidget(self.toolbar)
        self._central_layout.addWidget(self.client)

        # 超·布局
        self._super_layout = QVBoxLayout(self._super)
        self._super_layout.addWidget(self._central)
        self._super.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)

        self.setMinimumHeight(self.titlebar.title_label.height())

    def set_titlebar_height(self, height):
        """设置标题栏高度"""

        self.titlebar.setFixedHeight(height)

    def _adapt_bg(self, image):
        """返回适应窗口大小的背景图片"""
        image = image.scaled(self.width(), self.height(), \
                             Qt.KeepAspectRatioByExpanding, \
                             Qt.SmoothTransformation)
        return image

    def set_bg(self, image):
        """设置本地背景图片

        形参 image 可以是 QPixmap 实例、资源路径或磁盘路径。
        """
        if type(image) == QPixmap:
            self._bgimg = image
        elif type(image) == str:
            image = image.strip()
            if image[:4] == 'http':
                logging.warning("set_bg(self, image: Union[QPixmap, path])" +
                                "\n:  The argument 1 is a URL like, you might want to use" +
                                " set_web_bg(self, url, params=None, **kwargs).")
                self.set_web_bg(image)
                return None
            elif not image:
                logging.warning("set_bg(self, image: Union[QPixmap, path])" +
                                "\n:  The argument 1 is null.")
                return None
            elif not exists(image) and image[:2] != ':/':
                logging.warning("set_bg(self, image: Union[QPixmap, path])" +
                                "\n:  The file '{}' is inexistence.".format(image))
                return None
            image = QPixmap(image)
            self._bgimg = image
        else:
            return None
        adapt_image = self._adapt_bg(self._bgimg)
        palette = QPalette()
        palette.setBrush(QPalette.Window, QBrush(adapt_image))
        self.setPalette(palette)

    def set_web_bg(self, url, params=None, **kwargs):
        """使用 requests.get 加载来自网络的背景图片"""

        if url[:4] == 'http':
            try:
                response = get(url, params=None, **kwargs)
            except Exception as e:
                loggin.warning('Error: OcGui.OWindow.set_web_bg, ', e)
                return False
            else:
                img = QPixmap()
                img.loadFromData(response.content)
                self._bgimg = img
                palette = QPalette()
                adapt = self._adapt_bg(img)
                palette.setBrush(QPalette.Window, QBrush(adapt))
                self.setPalette(palette)
                return True

    def add_tools(self, widget):
        """添加工具栏"""

        if not isinstance(widget, QWidget):
            logging.warning('add_tools(self, widget: QWidget): argument 1 ' +
                            'is unexpected type {}'.format(type(widget)))
            return None

        return self.toolbar.addWidget(widget)

    def set_page(self, index, obj='client'):
        """切换页面"""

        if obj == 'client':
            self.client.setCurrentIndex(index)
        elif obj == 'toolbar':
            self.toolbar.setCurrentIndex(index)

    def _change_cursor_icon(self, area):
        """改变鼠标在窗口边缘时的图片"""

        if self.maximumWidth() == self.minimumWidth() and (area == 21 or area == 23):
            return None
        if self.maximumHeight() == self.minimumHeight() and (area == 12 or area == 32):
            return None

        if area == 11 or area == 33:
            self.setCursor(Qt.SizeFDiagCursor)
        elif area == 12 or area == 32:
            self.setCursor(Qt.SizeVerCursor)
        elif area == 13 or area == 31:
            self.setCursor(Qt.SizeBDiagCursor)
        elif area == 21 or area == 23:
            self.setCursor(Qt.SizeHorCursor)
        else:
            self.setCursor(Qt.ArrowCursor)

    def _compute_area(self, pos):
        """计算鼠标在窗口中的区域

        Args:
            pos: 鼠标相对于窗口的位置
            margin: 以此值为外边框宽度，划为九宫格区域
        """
        margin = self.MARGIN

        # 定位列坐标
        if pos.x() < margin:
            line = 1
        elif pos.x() > self.width() - margin:
            line = 3
        else:
            line = 2
        # 定位行坐标并结合列坐标
        if pos.y() < margin:
            return 10 + line
        elif pos.y() > self.height() - margin:
            return 30 + line
        else:
            return 20 + line

    def _pos_percent(self, pos):
        """返回鼠标相对窗口的纵横百分比"""

        if pos.x() <= 0:
            x = 0
        else:
            x = round(pos.x() / self.width(), 3)

        if pos.y() <= 0:
            y = 0
        else:
            y = round(pos.y() / self.height(), 3)

        return x, y

    def _move(self, event):
        """实现窗口移动"""

        self._move_count += 1

        # 判断移动次数，减少灵敏度
        if self._move_count < self.MOVE_SENSITIVITY:
            return None

        # 最大化时需恢复并移动到鼠标位置
        if self.isMaximized():
            relative = self._pos_percent(event.pos())
            self.titlebar.max_button_click()
            gpos = event.globalPos()
            width = self._last_geometry.width()
            height = self._last_geometry.height()
            x = gpos.x() - round(width * relative[0])
            y = gpos.y() - round(height * relative[1])

            self.setGeometry(x, y, width, height)
        else:
            # 鼠标移动偏移量
            offsetPos = event.globalPos() - self._posLast
            # ~ print(self.pos(), '->', self.pos() + offsetPos)
            self.move(self.pos() + offsetPos)

    def _resize(self, event):
        """实现拖动调整窗口大小的函数

        以新旧坐标差计算偏移量，使用 QRect 实例附带位置坐标；
        核心算法做了三重校验，以确保任意情况下窗口都能以正确的方式调整大小：
            一: 横纵坐标与最值校验，确保在最值范围内调整大小；
            二: 横纵坐标与左右区块校验，确保鼠标在窗口边缘时才调整大小；
            三: 横纵坐标与极值偏移量校验，确保在改变坐标的情况下，窗口不会发生漂移
        """
        # 鼠标在窗口中的区域
        area = self._area
        # 鼠标偏移量
        offsetPos = event.globalPos() - self._posLast
        # 鼠标在窗口中的坐标
        winPos = event.pos()

        # 矩形实例，被赋予窗口的几何属性（x, y, width, height）
        # 利用其改变左上角坐标，但右下角坐标不变的特性，实现窗口移动效果
        rect = QRect(self.geometry())

        x = rect.x()
        y = rect.y()
        width = rect.width()
        height = rect.height()

        minWidth = self.minimumWidth()
        minHeight = self.minimumHeight()
        maxWidth = self.maximumWidth()
        maxHeight = self.maximumHeight()

        # 根据不同区域选择不同操作
        if area == 11:
            # 左上
            pos = rect.topLeft()

            if offsetPos.x() < 0 and width < maxWidth or offsetPos.x() > 0 and width > minWidth:
                if offsetPos.x() < 0 and winPos.x() <= 0 or offsetPos.x() > 0 and winPos.x() >= 0:
                    if (maxWidth - width) >= -offsetPos.x() and (width - minWidth) >= offsetPos.x():
                        pos.setX(pos.x() + offsetPos.x())

            if offsetPos.y() < 0 and height < maxHeight or offsetPos.y() > 0 and height > minHeight:
                if offsetPos.y() < 0 and winPos.y() <= 0 or offsetPos.y() > 0 and winPos.y() >= 0:
                    if (maxHeight - height) >= -offsetPos.y() and (height - minHeight) >= offsetPos.y():
                        pos.setY(pos.y() + offsetPos.y())

            rect.setTopLeft(pos)

        elif area == 13:
            # 右上
            pos = rect.topRight()

            if offsetPos.x() < 0 and width > minWidth or offsetPos.x() > 0 and width < maxWidth:
                if offsetPos.x() < 0 and winPos.x() <= width or offsetPos.x() > 0 and winPos.x() >= width:
                    pos.setX(pos.x() + offsetPos.x())

            if offsetPos.y() < 0 and height < maxHeight or offsetPos.y() > 0 and height > minHeight:
                if offsetPos.y() < 0 and winPos.y() <= 0 or offsetPos.y() > 0 and winPos.y() >= 0:
                    if (maxHeight - height) >= -offsetPos.y() and (height - minHeight) >= offsetPos.y():
                        pos.setY(pos.y() + offsetPos.y())

            rect.setTopRight(pos)

        elif area == 31:
            # 左下
            pos = rect.bottomLeft()

            if offsetPos.x() < 0 and width < maxWidth or offsetPos.x() > 0 and width > minWidth:
                if offsetPos.x() < 0 and winPos.x() <= 0 or offsetPos.x() > 0 and winPos.x() >= 0:
                    if (maxWidth - width) >= -offsetPos.x() and (width - minWidth) >= offsetPos.x():
                        pos.setX(pos.x() + offsetPos.x())

            if offsetPos.y() < 0 and height > minHeight or offsetPos.y() > 0 and height < maxHeight:
                if offsetPos.y() < 0 and winPos.y() <= height or offsetPos.y() > 0 and winPos.y() >= height:
                    pos.setY(pos.y() + offsetPos.y())

            rect.setBottomLeft(pos)

        elif area == 33:
            # 右下
            pos = rect.bottomRight()

            if offsetPos.x() < 0 and width > minWidth or offsetPos.x() > 0 and width < maxWidth:
                if offsetPos.x() < 0 and winPos.x() <= width or offsetPos.x() > 0 and winPos.x() >= width:
                    pos.setX(pos.x() + offsetPos.x())

            if offsetPos.y() < 0 and height > minHeight or offsetPos.y() > 0 and height < maxHeight:
                if offsetPos.y() < 0 and winPos.y() <= height or offsetPos.y() > 0 and winPos.y() >= height:
                    pos.setY(pos.y() + offsetPos.y())

            rect.setBottomRight(pos)

        elif area == 12:
            # 中上
            if offsetPos.y() < 0 and height < maxHeight or offsetPos.y() > 0 and height > minHeight:
                if offsetPos.y() < 0 and winPos.y() <= 0 or offsetPos.y() > 0 and winPos.y() >= 0:
                    if (maxHeight - height) >= -offsetPos.y() and (height - minHeight) >= offsetPos.y():
                        rect.setTop(rect.top() + offsetPos.y())

        elif area == 21:
            # 中左
            if offsetPos.x() < 0 and width < maxWidth or offsetPos.x() > 0 and width > minWidth:
                if offsetPos.x() < 0 and winPos.x() <= 0 or offsetPos.x() > 0 and winPos.x() >= 0:
                    if (maxWidth - width) >= -offsetPos.x() and (width - minWidth) >= offsetPos.x():
                        rect.setLeft(rect.left() + offsetPos.x())

        elif area == 23:
            # 中右
            if offsetPos.x() < 0 and width > minWidth or offsetPos.x() > 0 and width < maxWidth:
                if offsetPos.x() < 0 and winPos.x() <= width or offsetPos.x() > 0 and winPos.x() >= width:
                    rect.setRight(rect.right() + offsetPos.x())

        elif area == 32:
            # 中下
            if offsetPos.y() < 0 and height > minHeight or offsetPos.y() > 0 and height < maxHeight:
                if offsetPos.y() < 0 and winPos.y() <= height or offsetPos.y() > 0 and winPos.y() >= height:
                    rect.setBottom(rect.bottom() + offsetPos.y())

        # 设置窗口几何属性（坐标，宽高）
        self.setGeometry(rect)

    # 重写方法
    def setWindowIcon(self, icon='default', width=16, height=16):
        """设置窗口图标"""

        if icon == 'default':
            icon = set_icon(self.titlebar._title_icon, ':/res/images/one_ccs.ico',
                            width=width, height=height)
        else:
            icon = set_icon(self.titlebar._title_icon, icon, width=width, height=height)

        QWidget.setWindowIcon(self, QIcon(icon))

    def setWindowTitle(self, string):
        """设置窗口标题"""
        self.titlebar.title_label.setText(string)

    def addWidget(self, widget):
        """每添加一个 widget 既一页"""

        if not isinstance(widget, QWidget):
            logging.warning('addWidget(self, widget: QWidget): argument 1 ' +
                            'is unexpected type {}'.format(type(widget)))
            return None

        return self.client.addWidget(widget)

    def addLayout(self, *args):
        """禁止添加布局"""
        logging.warning("You shouldn't useadd Layout(self, *args), " +
                        "you should use addWidget(self, widget).")

    def setStyleSheet(self, qss, default='enabled'):
        """加载样式表"""

        if not isinstance(qss, str):
            logging.warning('setStyleSheet(self, qss: str): argument 1' +
                            'is a unexpected type ' + type(qss))
            return None

        if default != 'disabled':
            qss = OcGuiRes.qss + qss

        if qss[:8] == 'easy_pyoc.qss':
            qss = qss[8:]
            QWidget.setStyleSheet(self, qss)
        elif qss.find('\\') or qss.find('/'):
            try:
                with open(qss, 'r') as f:
                    lines = f.readlines()
                    stylesheet = ''.join(lines)
            except Exception as e:
                logging.warning('Error of function OWindow.setStyleSheet()\n', e)
            else:
                QWidget.setStyleSheet(self, stylesheet)
        else:
            QWidget.setStyleSheet(self, qss)

    # 重写事件
    def eventFilter(self, obj, event):
        """事件过滤器"""

        minb = self.titlebar._min_button
        maxb = self.titlebar._max_button
        clob = self.titlebar._close_button
        skib = self.titlebar._skin_button
        menb = self.titlebar._menu_button

        if obj == minb or obj == maxb or obj == clob or obj == skib or obj == menb:
            self.setCursor(Qt.ArrowCursor)
            if obj == maxb and event.type() == QEvent.MouseButtonPress:
                self._last_geometry = self.geometry()

            # ~ return True  # 说明这个事件已被处理，其他控件别插手

        return QObject.eventFilter(self, obj, event)  # 交由其他控件处理

    def resizeEvent(self, event):
        """重写窗口大小改变事件，实现自适应背景图片"""

        if not self._bgimg:
            return None
        palette = QPalette()
        img = self._adapt_bg(self._bgimg)
        palette.setBrush(QPalette.Window, QBrush(img))
        self.setPalette(palette)

    def mouseDoubleClickEvent(self, event):
        """重写继承的鼠标双击事件"""

        self._last_geometry = self.geometry()
        self.titlebar.max_button_click()

        return QMainWindow.mouseDoubleClickEvent(self, event)

    def mousePressEvent(self, event):
        """重写继承的鼠标按住事件"""

        self._isPressed = True
        self._press_button = event.button()
        self._area = self._compute_area(event.pos())
        self._move_count = 0
        self._posLast = event.globalPos()

        return QMainWindow.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        """重写继承的鼠标释放事件"""

        self._isPressed = False
        self._press_button = None
        self._area = None
        self._move_count = 0
        self.setCursor(Qt.ArrowCursor)

        return QMainWindow.mouseReleaseEvent(self, event)

    def mouseMoveEvent(self, event):
        """重写继承的鼠标移动事件，实现拖动改变窗口大小"""

        area = self._compute_area(event.pos())

        # 调整窗口大小及移动
        if self._isPressed and self._press_button == Qt.LeftButton:
            if self._area == 22:
                self._move(event)
            elif not self.isMaximized():
                self._resize(event)

            # 更新鼠标全局坐标
            self._posLast = event.globalPos()
            return None
        if not self._isPressed and not self.isMaximized():
            # 调整鼠标图标，按下鼠标后锁定状态
            self._change_cursor_icon(area)

        return QMainWindow.mouseMoveEvent(self, event)


def set_icon(obj, icon, width=20, height=20):
    """为 QLabel 或 QPushButton 设置图像"""

    if not isinstance(obj, QLabel) and not isinstance(obj, QPushButton):
        logging.warning('set_icon(obj: Union[QLabel, QPushButton], icon,' +
                        ' width=20, height=20): argument 1 is unexpected' +
                        ' type {}'.format(type(obj)))
        return None

    KAR = Qt.KeepAspectRatio
    STFM = Qt.SmoothTransformation

    if isinstance(obj, QLabel):
        if isinstance(icon, str):
            icon = QPixmap(icon).scaled(width, height, KAR, STFM)

        obj.setPixmap(icon)
    else:
        if isinstance(icon, str):
            icon = QIcon(QPixmap(icon))
        if isinstance(icon, QPixmap):
            icon = QIcon(icon)

        obj.setIcon(icon)
        obj.setIconSize(QSize(width, height))

    return icon


def set_all_gap(layout, left=0, top=0, right=0, bottom=0, spacing=0):
    """设置布局间的所有间距"""

    # 设置布局边距(左,顶,右,底)
    layout.setContentsMargins(left, top, right, bottom)
    # 去除布局及控件间的间隔
    layout.setSpacing(spacing)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Union
from flask import Request

from .object_util import ObjectUtil


class FlaskUtil(object):

    @staticmethod
    def quick_data(request: Request, *keys) -> Union[tuple, dict, any]:
        """将 flask.Request 中的 args, values, form, files, json 解构为元组

        例 <json>::
            quick_data(request, 'id', 'coords.longitude', 'coords.latitude')
        例 <form>::
            quick_data(request, ('id', int), ('coords[longitude]', float), ('coords[latitude]',float))

        :param keys: None 直接返回整个数据 (dict 类型)
        :param keys: str 返回取到的值
        :param keys: (key: str, type: T) 索引 0 为字段名, 索引 1 为字段类型
        :param keys: (key: str, type: T, default: any) 索引 0 为字段名, 索引 1 为字段类型, 索引 2 为默认值
        :return -> (tuple | dict | any) 返回元组或字典, 当结果列表长度为 1 时直接返回
        """
        # 获取请求数据
        data = {
            **request.args.to_dict(),
            **request.values.to_dict(),
            **request.form.to_dict(),
            **request.files.to_dict(),
        }
        json_data = request.get_json(force=True, silent=True)
        if json_data:
            data.update(json_data)

        if not keys:
            return data
        # 解析为元组
        values = []
        for key in keys:
            if isinstance(key, str):
                values.append(ObjectUtil.get_value_from_dict(data, key))
            elif isinstance(key, list) or isinstance(key, tuple):
                if len(key) == 2:
                    value = ObjectUtil.get_value_from_dict(data, key[0])
                    values.append(key[1](value) if value else None)
                elif len(key) == 3:
                    value = ObjectUtil.get_value_from_dict(data, key[0], key[2])
                    values.append((key[1](value) if value != None else None))
        return values[0] if len(values) == 1 else tuple(values) if values else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Callable
import socket
import ipaddress

from ..sock.server_socket import ServerSocket
from ..sock.client_socket import ClientSocket


class NetworkUtil(object):

    @staticmethod
    def is_ip(ip_addr: str) -> bool:
        """判断是否为合法的 IP 地址

        Args:
            ip_addr (str): IP 地址

        Returns:
            bool: 是否为合法的 IP 地址
        """
        try:
            ipaddress.ip_address(ip_addr)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_host(host: str) -> bool:
        """判断是否为合法的主机名或 IP 地址

        Args:
            host (str): 主机名或 IP 地址

        Returns:
            bool: 是否为合法的主机名或 IP 地址
        """
        try:
            socket.gethostbyname(host)
            return True
        except socket.gaierror:
            return False

    @staticmethod
    def classify_ip(ip_addr: str) -> str:
        """根据 IP 地址的第一个字节判断 IP 地址的分类

        Args:
            ip_addr (str): IP 地址

        Raises:
            ValueError: 无效的 IP 地址

        Returns:
            str: IP 地址分类 {'A', 'B', 'C', 'D', 'E', None}
        """
        if not NetworkUtil.is_ip(ip_addr):
            raise ValueError('无效的 IP 地址')
        first_octet = int(ip_addr.split('.')[0])

        if 1 <= first_octet <= 126:
            return 'A'
        elif 128 <= first_octet <= 191:
            return 'B'
        elif 192 <= first_octet <= 223:
            return 'C'
        elif 224 <= first_octet <= 239:
            return 'D'
        elif 240 <= first_octet <= 255:
            return 'E'
        else:
            raise ValueError('无效的 IP 地址')

    @staticmethod
    def get_hostname() -> str:
        """返回一个字符串，包含当前正在运行 Python 解释器的机器的主机名。

        Returns:
            str: 主机名
        """
        return socket.gethostname()

    @staticmethod
    def get_host_by_name(hostname: str) -> str:
        """将主机名转换为 IPv4 地址格式。IPv4 地址以字符串格式返回，如 '100.50.200.5'。
        如果主机名本身是 IPv4 地址，则原样返回。

        Args:
            hostname (str): 主机名

        Returns:
            str: IP 地址
        """
        return socket.gethostbyname(hostname)

    @staticmethod
    def get_host_by_name_ex(hostname: str) -> tuple[str, list[str], list[str]]:
        """将一个主机名转换为 IPv4 地址格式的扩展接口。 返回一个 3 元组 (hostname, aliaslist, ipaddrlist)
        其中 hostname 是主机的首选主机名，aliaslist 是同一地址的备选主机名列表（可能为空），而 ipaddrlist
        是同一主机上同一接口的 IPv4 地址列表（通常为单个地址但并不总是如此）。

        Args:
            hostname (str): 主机名

        Returns:
            tuple[str, list[str], list[str]]: 3 元组 (hostname, aliaslist, ipaddrlist)
        """
        return socket.gethostbyname_ex(hostname)

    @staticmethod
    def get_local_ip(nth: int = 0) -> str:
        """获取本地 IP 地址

        Args:
            nth (int, optional): 多网卡时获取第 nth 个网卡的 IP 地址. 默认为 0.

        Raises:
            ValueError: 无效的网卡序号

        Returns:
            _type_: 本地 IP 地址
        """
        hostname = socket.gethostname()

        if nth == 0:
            ip = socket.gethostbyname(hostname)
        else:
            ips = socket.gethostbyname_ex(hostname)[2]
            if nth >= len(ips):
                raise ValueError(f'无效的网卡序号: {nth}')
            ip = ips[nth]
        return ip

    @staticmethod
    def get_local_ips() -> list[str]:
        """获取本地 IP 地址列表

        Returns:
            _type_: 本地 IP 地址列表
        """
        hostname = socket.gethostname()
        ips = socket.gethostbyname_ex(hostname)[2]
        return ips

    @staticmethod
    def get_broadcast_address(ip_network: str) -> str:
        """返回对应 IP 地址的广播地址

        Example:
            >>> NetworkUtil.get_broadcast_address('192.168.1.100/24')
            '192.168.1.255'

            >>> NetworkUtil.get_broadcast_address('192.168.1.100')
            '192.168.1.255'

        Args:
            ip_addr (str): IP 地址

        Returns:
            str: 广播地址
        """
        if not '/' in ip_network:
            if classify := NetworkUtil.classify_ip(ip_network):
                netmask = {'A': '/8', 'B': '/16', 'C': '/24', 'D': '/24', 'E': '/32'}[classify]
                ip_network = f'{ip_network}{netmask}'

        ip_network = ipaddress.ip_network(ip_network, strict=False)
        return str(ip_network.broadcast_address)

    @staticmethod
    def create_multicast_server(
        group: tuple[str, int],
        on_recv: Callable[[bytes, tuple[str, int], Callable[[bytes], int]], None],
    ) -> ServerSocket:
        """创建一个组播服务端

        Args:
            group (tuple[str, int]): 组播地址
            on_recv (Callable[[bytes, tuple[str, int], Callable[[bytes], int]], None]): 接收到数据时的回调函数

        Returns:
            ServerSocket: 组播服务端套接字
        """
        return ServerSocket(protocol='MULTICAST', bind=('', group[1]), group=group[0], on_recv=on_recv)

    @staticmethod
    def create_multicast_client(
        group: tuple[str, int],
        on_recv: Callable[[bytes, tuple[str, int]], None],
    ) -> ClientSocket:
        """创建一个组播客户端

        Args:
            group (tuple[str, int]): 组播地址
            on_recv (Callable[[bytes, tuple[str, int]], None]): 接收到数据时的回调函数

        Returns:
            ClientSocket: 组播客户端套接字
        """
        return ClientSocket(protocol='MULTICAST', target=group, on_recv=on_recv)

    @staticmethod
    def send_WOL(mac_hex: str, *, port: int = 9527) -> None:
        """向本机的所有网卡发送网络唤醒包

        Args:
            mac_hex (str): mac 地址
            port (int, optional): 发送广播的端口. 默认为 9527.
        """
        mac_hex      = mac_hex.replace(':', '').replace('-', '')
        magic_packet = bytes.fromhex('ff' * 6 + mac_hex * 16)

        for local_ip in NetworkUtil.get_local_ips():
            broadcast_host = NetworkUtil.get_broadcast_address(local_ip)

            client_socket = ClientSocket(protocol="UDP", target=(broadcast_host, port), bind=(local_ip, 0))
            client_socket.send(magic_packet)
            client_socket.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from easy_pyoc import Logger


def test_logger():
    logger1 = Logger()
    logger2 = Logger()
    logger3 = Logger(name='test', formatter='%(message)s')

    assert logger1 is logger2
    assert logger1 is not logger3
    assert logger2 is not logger3



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from traceback import format_exc


class ExceptionUtil(object):
    """异常工具栏"""

    @staticmethod
    def format_exc() -> str:
        """返回异常调用信息"""
        return format_exc()

    @staticmethod
    def get_message() -> str:
        """返回异常信息"""
        return format_exc().split('\n')[-2].strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


with open('README.md', 'r', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


setup(
    name='easy_pyoc',
    version='0.11.1',
    description='封装一些 Python 小工具，提升开发效率。',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='one-ccs',
    author_email='one-ccs@foxmail.com',
    url='https://github.com/one-ccs/easy_pyoc',
    repository='https://github.com/one-ccs/easy_pyoc.git',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=3',
    packages=find_packages(
        where='.',
        exclude=('test', ),
        include=('*', ),
    ),
    package_dir={},
    package_data={
        'easy_pyoc': ['qt5/res/**'],
    },
    exclude_package_data={},
    install_requires=[
        'typing_extensions',
    ],
    extras_require={
        'easy_pyoc.qt5': ['PyQt5'],
        'utils.flask_util': ['flask'],
        'utils.yaml_util': ['PyYAML'],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Callable, Literal
from threading import Thread
from multiprocessing import Process
import socket

from ..classes.logger import Logger


class ClientSocket():
    def __init__(
            self,
            *,
            protocol: Literal['TCP', 'UDP', 'MULTICAST'],
            target: tuple[str, int],
            bind: tuple[str, int] | None = None,
            on_recv: Callable[[bytes, tuple[str, int]], None] | None = None,
            timeout: float = 5.0,
            is_process: bool = False,
        ):
        """客户端套接字

        发送 TCP/UDP/MULTICAST 数据并接收响应。

        Args:
            protocol (str): 协议
            target (tuple[str, int]): 服务器地址和端口
            bind (tuple[str, int] | None, optional): 绑定地址, 端口为 `0` 时随机分配端口. 默认为 None.
            on_recv (Callable[[tuple[str, int], bytes], None] | None, optional): 接收到数据时的回调函数, 参数为 (数据, 地址). 默认为 None.
            timeout (float, optional): TCP 连接超时时间, 单位为秒. 默认为 5.0.
            is_process (bool, optional): 是否使用进程模式接收数据, 即在子进程中运行. 默认为 False.

        Raises:
            ValueError: 无效的协议类型, 应为 [TCP, UDP, MULTICAST]
            ValueError: 无效的端口号, 应为 [1-65535]
            ValueError: 无效的绑定端口号, 应为 [1-65535]
        """
        self.__active = False
        self.__socked = False

        if protocol not in ['TCP', 'UDP', 'MULTICAST']:
            raise ValueError(f'ClientSocket 无效的协议类型 "{protocol}"')
        if target[1] < 1 or target[1] > 65535:
            raise ValueError(f'ClientSocket 无效的端口号 "{target[1]}"')
        if bind and (bind[1] < 0 or bind[1] > 65535):
            raise ValueError(f'ClientSocket 无效的绑定端口号 "{bind[1]}"')
        if on_recv and not callable(on_recv):
            raise ValueError(f'ClientSocket on_recv 必须为可调用对象')

        self.logger     = Logger()
        self.protocol   = protocol
        self.target     = target
        self.bind       = bind
        self.on_recv    = on_recv
        self.timeout    = timeout
        self.is_process = is_process
        self.sock: socket.socket | None = None
        self.thread: Thread | Process | None = None

    def __str__(self) -> str:
        if self.bind:
            return f'ClientSocket({self.protocol}, {self.target[0]}:{self.target[1]}, bind {self.bind[0]}:{self.bind[1]})'
        return f'ClientSocket({self.protocol}, {self.target[0]}:{self.target[1]})'

    def __del__(self):
        self.close()

    def __create_socket(self) -> bool:
        if self.__socked:
            return True

        try:
            match self.protocol:
                case 'TCP':
                    self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    if self.bind:
                        self.sock.bind(self.bind)
                    self.sock.settimeout(self.timeout)
                    self.sock.connect((self.target[0], self.target[1]))
                case 'UDP':
                    self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    if self.bind:
                        self.sock.bind(self.bind)
                    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    if self.target[0].endswith('.255'): # 设置 SO_BROADCAST 为 1, 允许发送广播数据包
                        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                case 'MULTICAST':
                    self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
                    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
                    self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 1)
                    if self.bind:
                        self.sock.bind(self.bind)
                    self.sock.setsockopt(
                        socket.IPPROTO_IP,
                        socket.IP_ADD_MEMBERSHIP,
                        socket.inet_aton(self.target[0]) + socket.inet_aton('0.0.0.0'),
                    )
            if not self.bind or not self.bind[1]:
                self.bind = self.sock.getsockname()

            self.__socked = True
        except ConnectionRefusedError:
            self.logger.warning(f'{self} 无法连接: {self.target[0]}:{self.target[1]}')
        except Exception as e:
            self.logger.error(f'{self} 创建失败: \n{e}')

        return self.__socked

    def __recv_thread(self) -> None:
        self.__active = True

        while self.__active:
            try:
                data, addr = self.sock.recvfrom(1024)

                self.logger.debug(f'{self} 收到 {addr} 的数据: {data}')
                try:
                    self.on_recv(data, addr)
                except Exception as e:
                    self.logger.error(f'{self} "on_recv" 回调函数发生异常: \n{e}')
            except OSError as e:
                if e.errno == 10057:
                    self.logger.debug(f'{self} 接收失败连接未建立')
                    break
                elif e.errno == 10038: # 调用了 close() 方法
                    pass
                else:
                    self.logger.error(f'{self} 接收失败: \n{e}')

    def getpeername(self) -> tuple[str | None, int | None]:
        """返回套接字连接到的远程地址。"""
        if self.__create_socket():
            return self.sock.getpeername()
        return (None, None)

    def getsockname(self) -> tuple[str | None, int | None]:
        """返回套接字本身的地址。"""
        if self.__create_socket():
            return self.sock.getsockname()
        return (None, None)

    def send(self, data: bytes) -> None:
        if not self.__create_socket():
            self.logger.warning(f'{self} 未连接, 无法发送数据')
            return

        try:
            if self.protocol == 'TCP':
                self.sock.sendall(data)
            else:
                self.sock.sendto(data, (self.target[0], self.target[1]))
            self.logger.debug(f'{self} 发送数据: {data}')

            if self.on_recv and not self.thread:
                if self.is_process:
                    self.thread = Process(target=self.__recv_thread, daemon=True)
                else:
                    self.thread = Thread(target=self.__recv_thread, daemon=True)
                self.thread.start()
        except OSError as e:
            if e.errno == 10057:
                self.logger.debug(f'{self} 发送失败连接未建立')
            else:
                self.logger.error(f'{self} 发送失败: \n{e}')

    def close(self) -> bool:
        if self.__socked:
            try:
                self.__active = False
                self.sock.shutdown(socket.SHUT_RDWR)
                self.sock.close()
                self.__socked = False
                self.logger.debug(f'{self} 已关闭')
            except Exception as e:
                self.logger.error(f'{self} 关闭失败: \n{e}')
        return not self.__socked



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, TypeVar, Optional
from pathlib import Path
from contextlib import contextmanager
import sys
import os


if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath


T = TypeVar('T')


class PathUtil(object):
    """路径工具类"""

    Path = Path
    sys_path = sys.path

    @staticmethod
    def get_env(key: str, default: Optional[T] = None) -> str | T:
        """获取环境变量"""
        return os.getenv(key, default)

    @staticmethod
    def get_work_dir() -> str:
        """获取工作目录"""
        return os.getcwd()

    @staticmethod
    def set_work_dir(path: 'FileDescriptorOrPath') -> None:
        """设置工作目录"""
        os.chdir(path)

    @staticmethod
    def get_home_dir() -> Path:
        """获取用户目录"""
        return Path.home()

    @staticmethod
    def get_dir(path: str) -> Path:
        """获取路径的目录"""
        p = Path(path)
        return p if p.is_dir() else p.parent

    @staticmethod
    def abspath(path: str = '') -> str:
        """返回路径的绝对路径"""
        return str(Path(path).absolute())

    @staticmethod
    def is_exists_file(path: str) -> bool:
        """判断文件是否存在"""
        _path = Path(path)
        return _path.exists() and _path.is_file()

    @staticmethod
    def is_exists_dir(path: str) -> bool:
        """判断路径是否是个存在"""
        _path = Path(path)
        return _path.exists() and _path.is_dir()

    @staticmethod
    @contextmanager
    def open(fp: 'FileDescriptorOrPath', mode: str, buffering: int = -1, encoding: str = 'utf-8'):
        """打开文件"""
        try:
            with open(fp, mode=mode, buffering=buffering, encoding=encoding) as f:
                yield f
        except FileNotFoundError:
            raise FileNotFoundError(f'文件 "{PathUtil.abspath(fp)}" 不存在')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Callable, TypeVar, ParamSpecArgs, ParamSpecKwargs, Any
from traceback import format_exc
from functools import wraps
from inspect import signature


_T = TypeVar('_T')


class FuncUtil(object):
    """函数工具类"""

    @staticmethod
    def is_arg(func: Callable, name: str) -> bool:
        """判断函数是否有指定名称的位置参数

        Args:
            func (Callable): 函数
            name (str): 参数名称

        Returns:
            bool: 是否有指定名称的位置参数
        """
        sig = signature(func)

        if param := sig.parameters.get(name):
            return param.kind == param.POSITIONAL_ONLY or '=' not in str(param)

        return False

    @staticmethod
    def is_kwarg(func: Callable, name: str) -> bool:
        """判断函数是否有指定名称的关键字参数

        Args:
            func (Callable): 函数
            name (str): 参数名称

        Returns:
            bool: 是否有指定名称的关键字参数
        """
        sig = signature(func)

        if param := sig.parameters.get(name):
            return param.kind == param.KEYWORD_ONLY or '=' in str(param)

        return False

    @staticmethod
    def log(logger: Callable = print):
        """装饰器，打印函数调用信息

        Args:
            logger (Callable, optional): 日志输出函数. 默认为 `print`.
        """
        def decorator(func: Callable):
            @wraps(func)
            def wrapper(*args, **kwargs):
                msg = f'调用 "{func.__name__}" 函数'
                if args:   msg += f', 位置参数: {args}'
                if kwargs: msg += f', 关键字参数: {kwargs}'

                logger(msg)
                return func(*args, **kwargs)
            return wrapper
        return decorator

    @staticmethod
    def catch(logger: Callable = print, *, is_raise: bool = False, on_except: Callable | None = None):
        """装饰器，捕获函数异常并打印日志

        Args:
            logger (Callable, optional): 日志输出函数. 默认为 `print`.
            is_raise (bool, optional): 是否抛出异常. 默认为 False.
            on_except (Callable | None, optional): 异常回调函数. 默认为 None.
        """
        def decorator(func: Callable):
            @wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    fmt_exc = format_exc()
                    _, exc = fmt_exc.split('\n')[-3:-1]

                    call_fn, raise_fn = func.__name__, 'unknown'

                    if _.strip().startswith('File "<string>"'):
                        raise_fn = _.split(' ')[-1]
                        line      = _.split(' ')[-3][:-1]
                        msg = f'函数 {raise_fn} ({line})'
                    else:
                        msg = f'函数 {call_fn}'

                    msg += f', 发生异常: {exc}'
                    msg += f', 调用者: {call_fn}'

                    if args:   msg += f', 位置参数: {args}'
                    if kwargs: msg += f', 关键字参数: {kwargs}'

                    tb = f'错误回溯开始 {call_fn} -> {raise_fn} ' + '+' * 32 + '\n'
                    if args:   tb += f'位置参数: {args}\n'
                    if kwargs: tb += f'关键字参数: {kwargs}\n'
                    tb += f'{fmt_exc}\n'
                    tb += f'错误回溯结束 {call_fn} -> {raise_fn} ' + '+' * 32 + '\n'
                    logger(msg, tb)

                    if callable(on_except): on_except(e)
                    if is_raise: raise
            return wrapper
        return decorator

    @staticmethod
    def hook(hook: Callable[[str, ParamSpecArgs, ParamSpecKwargs], bool | Any]):
        """装饰器，在函数调用前后执行 hook 函数

        Args:
            hook (Callable[[str, ParamSpecArgs, ParamSpecKwargs], bool  |  Any]):
                钩子函数, 将调用两次.
                第一次参数为 ('before', None, *args, **kwargs), 返回 `...` 则跳过函数调用,
                第二次参数为 ('after', result, *args, **kwargs), 返回非 None 值将作为函数调用结果返回.
        """
        def decorator(func: Callable):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if callable(hook) and hook('before', None, *args, **kwargs) is ...:
                    return None

                try:
                    result = func(*args, **kwargs)
                finally:
                    if callable(hook) and (_ := hook('after', result, *args, **kwargs)) is not None:
                        return _

                return result
            return wrapper
        return decorator

    @staticmethod
    def call(target: Callable[..., _T], *args, **kwargs) -> _T:
        """调用函数，并忽略多余参数

        Args:
            target (Callable[..., _T]): 目标函数
            *args: 位置参数
            **kwargs: 关键字参数

        Returns:
            _T: 目标函数返回值
        """
        sig = signature(target)

        _args = 0
        _kwargs = {}
        for param in sig.parameters.values():
            if FuncUtil.is_kwarg(target, param.name):
                _kwargs[param.name] = kwargs.get(param.name, param.default)
            else:
                _args += 1
        _args = args[:_args]

        return target(*_args, **_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
from select import select
from dataclasses import dataclass
import socket

from .string_util import StringUtil
from .network_util import NetworkUtil


class KNXUtil(object):
    """KNXnet/IP 工具类"""

    KNX_PORT      = 3671
    """KNXnet/IP 端口号"""
    KNX_MULTICAST = ('224.0.23.12', KNX_PORT)
    """KNXnet/IP 多播地址"""
    KNX_MX        = 2
    """KNXnet/IP 多播 TTL"""
    H_S_REQ_EXT   = b'\x02\x0b'
    """KNXnet/IP 发现请求扩展头"""
    H_S_REQ       = b'\x02\x01'
    """KNXnet/IP 发现请求头"""
    H_S_RES       = b'\x02\x02'
    """KNXnet/IP 发现响应头"""
    DIBS          = b'\x08\x04\x01\x02\x08\x06\x07\x00'
    """KNXnet/IP 发现请求 DIBS"""

    @dataclass
    class ControlDevice(object):
        """KNX 网关设备信息

        Attributes:
            host (str): 网关设备 IP 地址
            port (int): 网关设备端口号
            inad (tuple[int, int, int]): 网关设备地址
            sn (str): 网关设备序列号
            mac (str): 网关设备 MAC 地址
            name (str): 网关设备名称
        """
        host: str
        port: int
        inad: tuple[int, int, int]
        sn: str
        mac: str
        name: str

    @staticmethod
    def discover(timeout: int = 5) -> list[ControlDevice]:
        """发现 KNX 网关设备

        Args:
            timeout (int, optional): 发现超时时间. 默认为 5.

        Returns:
            list[KNXUtil.ControlDevice]: 网关设备列表
        """
        devices = {}
        for res in _scan(timeout):
            if res in devices:
                continue

            mac = res[32: 38].hex()
            device = {
                'host': StringUtil.hex_to_ip(res[8: 12].hex()),
                'port': StringUtil.str_to_int(res[12: 14].hex(), 16),
                'inad': (res[18] & 0x0f, res[18] >> 4 & 0x0f, res[19]),
                'sn':   f'{res[22: 24].hex()}:{res[24: 28].hex()}',
                'mac':  ':'.join([mac[i:i+2] for i in range(0, len(mac), 2)]),
                'name': res[38:12+res[14]].decode(),
            }
            devices[mac] = KNXUtil.ControlDevice(**device)

        return list(devices.values())


def _make_header(identifier: bytes, length: int):
    return b'\x06\x10' + identifier + (length).to_bytes(2, 'big')


def _make_hpai(ip: str, port: int) -> bytes:
    return b'\x08\x01' + bytes.fromhex(
        StringUtil.ip_to_hex(ip) +
        StringUtil.int_to_str(port, 16, 4)
    )


def _scan(timeout: int = 5):
    stop_wait = datetime.now() + timedelta(seconds=timeout)
    h_s_req_ext = _make_header(KNXUtil.H_S_REQ_EXT, 22)
    h_s_req     = _make_header(KNXUtil.H_S_REQ, 14)

    sockets: list[socket.socket] = []
    for ip in NetworkUtil.get_local_ips():
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, KNXUtil.KNX_MX)
            sock.bind((ip, 0))
            sockets.append(sock)
        except socket.error:
            pass

    for sock in sockets:
        hpai_de = _make_hpai(*sock.getsockname())
        knx_s_reqs = [h_s_req_ext + hpai_de + KNXUtil.DIBS, h_s_req + hpai_de]
        try:
            for req in knx_s_reqs:
                sock.sendto(req, KNXUtil.KNX_MULTICAST)
            sock.setblocking(False)
        except socket.error:
            sockets.remove(sock)
            sock.close()

    responses = []
    try:
        while sockets:
            time_diff = stop_wait - datetime.now()
            seconds_left = time_diff.total_seconds()
            if seconds_left <= 0:
                break

            ready = select(sockets, [], [], seconds_left)[0]

            for sock in ready:
                try:
                    data, addr = sock.recvfrom(1024)
                except socket.error:
                    sockets.remove(sock)
                    sock.close()
                    continue

                if data.startswith(b'\x06\x10\x02\x02'):
                    responses.append(data)
    finally:
        for sock in sockets:
            sock.close()

    return set(responses)
